/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.addons.netviewer;

import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.visualization.VisualizationViewer;
import edu.uci.ics.jung.visualization.control.AbstractPopupGraphMousePlugin;
import java.awt.Color;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.AbstractAction;
import javax.swing.JMenu;
import javax.swing.JPopupMenu;
import monalisa.resources.ResourceManager;
import monalisa.resources.StringResources;

/**
 * All PopUp Menus for the Netviewer
 * Mouse Popup Menu
 */
public class GraphPopupMousePlugin extends AbstractPopupGraphMousePlugin implements MouseListener {

    public static final String NORMAL = "NORMAL"; // Normal mouse mode
    public static final String SINGLE_EDGE = "SINGLE_EDGE"; // right click to create a new edge
    public static final String EDGE = "EDGE"; // add new edge mode (constant)
    public static final String DOUBLE_EDGE = "DOUBLE_EDGE"; // add new edge mode (constant)
    public static final String IN_VERTEX = "IN_VERTEX";
    public static final String OUT_VERTEX = "OUT_VERTEX";
    public static final String DELETE = "DELTE";
    public static final String DELETE_BEND = "DELETE_BEND";
    public static final String ADD_BEND = "ADD_BEND";
    public static final String ALIGN_X = "ALIGN_X";
    public static final String ALIGN_Y = "ALIGN_Y";

    private List<NetViewerNode> knockOutNodes = new ArrayList<NetViewerNode>();

    private NetViewer nv;
    private VisualizationViewer<NetViewerNode, NetViewerEdge> vv;
    private Graph<NetViewerNode, NetViewerEdge> g;

    private static final ResourceManager resources = ResourceManager.instance();
    private static final StringResources strings = resources.getDefaultStrings();

    private String mouseMode = NORMAL;
    private NetViewerNode source;

    private Set<NetViewerNode> pickedVertices = new HashSet<NetViewerNode>();
    private List<NetViewerNode> pickedVerticesList = new ArrayList<NetViewerNode>();
    private Set<NetViewerEdge> pickedEdges = new HashSet<NetViewerEdge>();
    private List<NetViewerEdge> pickedEdgesList = new ArrayList<NetViewerEdge>();

    public GraphPopupMousePlugin(NetViewer nv) {
        this.nv = nv;
        this.vv = nv.vv;
        this.g = nv.g;
    }

    public void changeGraph(Graph<NetViewerNode, NetViewerEdge> g) {
        this.g = g;
    }

    @Override
    protected void handlePopup(final MouseEvent me) {
        JPopupMenu popup = new JPopupMenu();

        pickedVertices = vv.getPickedVertexState().getPicked();
        pickedVerticesList.clear();
        for(NetViewerNode n : pickedVertices)
            pickedVerticesList.add(n);
        pickedEdges = vv.getPickedEdgeState().getPicked();
        pickedEdgesList.clear();
        for(NetViewerEdge edge : pickedEdges)
            pickedEdgesList.add(edge);
        int numberOfPickedVertices = pickedVertices.size();
        int numberOfPickedEdges = pickedEdges.size();
        boolean bendedEdge = true;
        
        if(this.mouseMode.equals(NORMAL)) {

            // Mutli Vertex Popup Menu
            if(numberOfPickedVertices > 1) {

                // Vertext Setup
                popup.add(new AbstractAction(strings.get("NVProperties")) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        nv.showVertexSetup(pickedVerticesList, me.getX(), me.getY());
                    }
                });

                Boolean meltLogicalPlaces = false;
                NetViewerNode firstNode = pickedVerticesList.get(0);
                if(firstNode.isLogical()) {
                    for(NetViewerNode nvNode : pickedVerticesList) {
                        if(nvNode.getNodeType().equalsIgnoreCase(NetViewer.PLACE)) {
                            if(nvNode.isLogical() && !nvNode.isMasterNode() && firstNode.getMasterNode().equals(nvNode.getMasterNode()))
                                meltLogicalPlaces = true;
                            else
                                meltLogicalPlaces = false;
                        }
                    }
                }

                // Melt logical places
                if(meltLogicalPlaces) {
                    popup.add(new AbstractAction(strings.get("NVMeltLogicalPlaces")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.removeLogicalPlaces(pickedVerticesList);
                        }
                    });
                }

                popup.addSeparator();

                // Align X
                popup.add(new AbstractAction(strings.get("NVAlignXPopUp")) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        nv.alignXMouseAction();
                    }
                });

                // Align Y
                popup.add(new AbstractAction(strings.get("NVAlignYPopUp")) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        nv.alignYMouseAction();
                    }
                });
                
                popup.addSeparator();
                
                // Export Subgraph
                popup.add(new AbstractAction(strings.get("NVExportSubGraph")) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        try {
                            nv.exportSelectedSubGraphMouseAction();
                        } catch (FileNotFoundException ex) {
                            Logger.getLogger(GraphPopupMousePlugin.class.getName()).log(Level.SEVERE, null, ex);
                        }
                    }
                });                   
            }

            // Edge Popup Menu
            if(numberOfPickedEdges > 1) {
                List<NetViewerEdge> bendList = pickedEdgesList.get(0).getMasterEdge().getBendEdges();
                for (NetViewerEdge picked : pickedEdges) {
                    if (!bendList.contains(picked)) {
                        bendedEdge = false;
                        break;
                    }
                }
            } 
            else {
                bendedEdge = false;
            }

            if( (numberOfPickedEdges == 1 || bendedEdge)  && numberOfPickedVertices == 0) {
                final NetViewerEdge edge = pickedEdgesList.get(0);

                // Edge Setup
                popup.add(new AbstractAction(strings.get("NVProperties")) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        nv.showEdgeSetup(edge, me.getX(), me.getY());
                    }
                });

                popup.addSeparator();

                // Insert Bend
                popup.add(new AbstractAction(strings.get("NVInsertBend")) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        nv.insertBend(me.getX(), me.getY());
                    }
                });

                // Remove Bend
                Boolean removeEdge = (edge.getAim().getNodeType().equals(NetViewer.BEND) || edge.getSource().getNodeType().equals(NetViewer.BEND)) ? false : true;
                if(!removeEdge) {
                    popup.add(new AbstractAction(strings.get("NVRemoveBend")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.removeBend();
                        }
                    });
                }

                // Plane Edge
                if(!removeEdge) {
                    popup.add(new AbstractAction(strings.get("NVPlaneEdge")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.planeEdge(edge, true);
                        }
                    });
                }

                popup.addSeparator();

                // Delete Edge
                popup.add(new AbstractAction(strings.get("NVDeleteEdge")) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        nv.deleteEdge(edge, false);
                    }
                });
            }

            // Vertex Popup Menu
            else if(numberOfPickedEdges == 0 && numberOfPickedVertices == 1) {
                final NetViewerNode node = pickedVerticesList.get(0);

                if(node.getNodeType().equals(NetViewer.BEND))
                    return;

                // Vertext Setup
                popup.add(new AbstractAction(strings.get("NVProperties")) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        nv.showVertexSetup(node, me.getX(), me.getY());
                    }
                });

                // Hide Label
                String itemName = (node.showLabel()) ? strings.get("NVHideLabel") : strings.get("NVShowLabel");
                popup.add(new AbstractAction(itemName) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        nv.hideLabel(node);
                    }
                });

                popup.addSeparator();

                if(node.getNodeType().equals(NetViewer.PLACE)) {
                    // Add Transition
                    JMenu addTransitionMenu = new JMenu(strings.get("NVCreateTransition"));

                    addTransitionMenu.add(new AbstractAction(strings.get("NVIn")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.createNewNode(node, true);
                        }
                    });

                    addTransitionMenu.add(new AbstractAction(strings.get("NVOut")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.createNewNode(node, false);
                        }
                    });

                    popup.add(addTransitionMenu);

                    // Delete Place
                    popup.add(new AbstractAction(strings.get("NVDeleteVertex")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.deleteNode(node, false);
                        }
                    });
                }
                else if(node.getNodeType().equals(NetViewer.TRANSITION)) {
                    // Add Place
                    JMenu addPlaceMenu = new JMenu(strings.get("NVCreateVertex"));

                    addPlaceMenu.add(new AbstractAction(strings.get("NVIn")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.createNewNode(node, true);
                        }
                    });

                    addPlaceMenu.add(new AbstractAction(strings.get("NVOut")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.createNewNode(node, false);
                        }
                    });

                    popup.add(addPlaceMenu);

                    // Delete Transition
                    popup.add(new AbstractAction(strings.get("NVDeleteTransition")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.deleteNode(node, false);
                        }
                    });
                }

                // Insert Edge
                popup.add(new AbstractAction(strings.get("NVCreateEdge")) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        source = node;
                        setMouseModeToSingleEdge();
                    }
                });

                // Create Locical Place
                Boolean addLocicalPlace = (node.getNodeType().equals(NetViewer.PLACE) && g.getNeighborCount(node) > 1) ? true : false;
                if(addLocicalPlace) {
                    popup.add(new AbstractAction(strings.get("NVCreateNewNodeButton")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.showCreateLogicalPlace(node, me.getX(), me.getY());
                        }
                    });
                }

                // Reverse Transition
                Boolean reverseTransition = (node.getNodeType().equals(NetViewer.TRANSITION)) ? true : false;
                if(reverseTransition) {
                    popup.add(new AbstractAction(strings.get("NVReverseTransition")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.createReverseTransition(node, me.getX(), me.getY());
                        }
                    });
                }

                // Delete Locical Place
                Boolean removeLocicalPlace = (node.getNodeType().equals(NetViewer.PLACE) && node.isLogical() && !node.isMasterNode()) ? true : false;
                if(removeLocicalPlace) {
                    popup.add(new AbstractAction(strings.get("NVRemoveNodeButton")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.removeLogicalPlace();
                        }
                    });

                    popup.add(new AbstractAction(strings.get("NVMeltAllLogicalPlaces")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            List<NetViewerNode> allPlaces = new ArrayList<NetViewerNode>();
                            for(NetViewerNode n : node.getMasterNode().getLogicalPlaces())
                                allPlaces.add(n);
                            nv.removeLogicalPlaces(allPlaces);
                        }
                    });
                }

                // Transition knock out
                JMenu knockOutMenu = new JMenu(strings.get("NVKnockoutTransition"));

                Boolean knockOutTransition = (node.getNodeType().equals(NetViewer.TRANSITION) && nv.hasTinvs()) ? true : false;
                if(knockOutTransition) {
                    popup.addSeparator();

                    knockOutMenu.add(new AbstractAction(strings.get("NVKnockOutStandart")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            knockOutNodes.add(node);
                            nv.knockOut(knockOutNodes, 0);
                        }
                    });

                    knockOutMenu.add(new AbstractAction(strings.get("NVKnockOutByOccurrence")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            knockOutNodes.add(node);
                            nv.knockOut(knockOutNodes, 1);
                        }
                    }).setEnabled(knockOutTransition);

                    knockOutMenu.add(new AbstractAction(strings.get("NVKnockOutByFactor")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            knockOutNodes.add(node);
                            nv.knockOut(knockOutNodes, 2);
                        }
                    });

                    knockOutMenu.add(new AbstractAction(strings.get("NVResetKnockOut")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            knockOutNodes.clear();
                            nv.resetKnockOut();
                        }
                    });
                    popup.add(knockOutMenu);
                }

                popup.addSeparator();

                // Align X
                popup.add(new AbstractAction(strings.get("NVAlignXPopUp")) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        nv.alignXMouseAction();
                    }
                });

                // Align Y
                popup.add(new AbstractAction(strings.get("NVAlignYPopUp")) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        nv.alignYMouseAction();
                    }
                });                                                            
            }

            // Generell Options
            else if(numberOfPickedEdges == 0 && numberOfPickedVertices == 0) {

                // MouseMode
                if(nv.getMouseMode()) {
                    popup.add(new AbstractAction(strings.get("NVGMTransforming")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.changeMouseModeToTransforming();
                        }
                    });
                }
                else {
                    popup.add(new AbstractAction(strings.get("NVGMPicking")) {
                        @Override
                        public void actionPerformed(ActionEvent e) {
                            nv.changeMouseModeToPicking();
                        }
                    });
                }

                popup.add(new AbstractAction(strings.get("NVCenterPetriNet")) {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        nv.center();
                    }
                });
            }
        }

        popup.show(vv, me.getX(), me.getY());
    }

    /**
     * Handles the picking of a new node
     * @param clickedNode
     */
    public void handlePicking(NetViewerNode clickedNode) {
        if(this.mouseMode.equals(EDGE) || this.mouseMode.equals(SINGLE_EDGE)) {
            if(!clickedNode.getNodeType().equalsIgnoreCase(source.getNodeType()) && !clickedNode.getNodeType().equals(NetViewer.BEND)) {
                if(!nv.createNewEdge(source, clickedNode)) {
                    nv.cancelMouseAction();
                    nv.setMessageLabelTo(strings.get("NVEdgeAlwaysHere"), Color.RED);
                    return;
                }
                nv.edgeMouseAction();
            }
            else {
                nv.setMessageLabelTo("<html>"+strings.get("NVInsertEdgeMessage") + "<br>" + strings.get("NVFalseAimType")+"</html>", Color.BLACK);
            }
        }
        else if(this.mouseMode.equals(DOUBLE_EDGE)) {
            this.source = clickedNode;            
            setMouseModeToEdge();
        }
        else if(this.mouseMode.equals(IN_VERTEX)) {
            if(!clickedNode.getNodeType().equals(NetViewer.BEND))
                nv.createNewNode(clickedNode, true);
        }
        else if(this.mouseMode.equals(OUT_VERTEX)) {
            if(!clickedNode.getNodeType().equals(NetViewer.BEND))
                nv.createNewNode(clickedNode, false);
        }
        else if(this.mouseMode.equals(DELETE)) {
            if(!clickedNode.getNodeType().equalsIgnoreCase(NetViewer.BEND))
                nv.deleteNode(clickedNode, false);
        }
        else if(this.mouseMode.equals(ALIGN_X)) {
            if (!clickedNode.getNodeType().equalsIgnoreCase(NetViewer.BEND))
                nv.alignVertices("X", clickedNode);
        }
        else if(this.mouseMode.equals(ALIGN_Y)) {
            if (!clickedNode.getNodeType().equalsIgnoreCase(NetViewer.BEND))
                nv.alignVertices("Y", clickedNode);
        }
    }

    /**
     * Handles the picking of a new edge
     * @param clickedEdge
     */
    public void handlePicking(NetViewerEdge clickedEdge) {
        if(this.mouseMode.equals(DELETE)) {
            nv.deleteEdge(clickedEdge, false);
        }
        else if(this.mouseMode.equals(ADD_BEND)) {
            Point mousePoint = vv.getMousePosition();
            nv.insertBend(clickedEdge, mousePoint.x, mousePoint.y, false);
        }
        else if(this.mouseMode.equals(DELETE_BEND)) {
            if(clickedEdge.getAim().getNodeType().equalsIgnoreCase(NetViewer.BEND) || clickedEdge.getSource().getNodeType().equalsIgnoreCase(NetViewer.BEND))
                nv.removeBend(clickedEdge, false);
        }
     }

    public void setMouseModeToNormal() {
        this.mouseMode = NORMAL;
    }

    public void setMouseModeToEdge() {
        this.mouseMode = EDGE;
        nv.setMessageLabelTo(strings.get("NVInsertEdgeMessage"), Color.BLACK);
    }

    public void setMouseModeToDoubleEdge() {
        this.mouseMode = DOUBLE_EDGE;
    }

    public void setMouseModeToInVertex() {
        this.mouseMode = IN_VERTEX;
    }

    public void setMouseModeToOutVertex() {
        this.mouseMode = OUT_VERTEX;
    }

    public void setMouseModeToDelete() {
        this.mouseMode = DELETE;
    }

    public void setMouseModeToDeleteBend() {
        this.mouseMode = DELETE_BEND;
    }

    public void setMouseModeToAddBend() {
        this.mouseMode = ADD_BEND;
    }

    private void setMouseModeToSingleEdge() {
        this.mouseMode = SINGLE_EDGE;
        nv.setMessageLabelTo(strings.get("NVInsertEdgeMessage"), Color.BLACK);
    }
    
    public void setMouseModeToAlignX() {
        this.mouseMode = ALIGN_X;
    }
        
    public void setMouseModeToAlignY() {
        this.mouseMode = ALIGN_Y;
    }

    public String getMouseMode() {
        return this.mouseMode;
    }

}
