/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.addons.netviewer;

import edu.uci.ics.jung.algorithms.layout.*;
import edu.uci.ics.jung.graph.*;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.EdgeType;
import edu.uci.ics.jung.visualization.*;
import edu.uci.ics.jung.visualization.control.ModalGraphMouse;
import edu.uci.ics.jung.visualization.decorators.EdgeShape;
import edu.uci.ics.jung.visualization.layout.PersistentLayout;
import edu.uci.ics.jung.visualization.layout.PersistentLayoutImpl;
import edu.uci.ics.jung.visualization.renderers.Renderer.VertexLabel.Position;
import org.apache.commons.collections15.functors.ConstantTransformer;

import java.io.FileNotFoundException;
import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.List;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ItemListener;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.awt.event.ItemEvent;
import java.awt.event.KeyEvent;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Stack;
import javax.swing.*;
import javax.imageio.ImageIO;
import layout.TableLayout;

import org.apache.batik.svggen.SVGGraphics2D;
import org.apache.batik.dom.GenericDOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.DOMImplementation;

import monalisa.Project;
import monalisa.Settings;
import monalisa.addons.topological.NodesDegrees;
import monalisa.resources.ResourceManager;
import monalisa.resources.StringResources;
import monalisa.data.pn.Arc;
import monalisa.data.pn.PInvariant;
import monalisa.data.pn.PetriNet;
import monalisa.data.pn.Place;
import monalisa.data.pn.Transition;
import monalisa.data.pn.TInvariant;
import monalisa.gui.MainDialog;
import monalisa.results.Configuration;
import monalisa.results.Mcts;
import monalisa.results.PInvariants;
import monalisa.results.PInvariantsConfiguration;
import monalisa.results.Result;
import monalisa.results.TInvariants;
import monalisa.results.TInvariantsConfiguration;
import monalisa.tools.Tool;
import monalisa.tools.tinv.TInvariantTool;
import monalisa.tools.mcts.MctsTool;
import monalisa.tools.pinv.PInvariantTool;
import monalisa.util.FileUtils;
import monalisa.addons.util.MyFileFilter;
import monalisa.addons.util.MyOutputFileFilter;
import monalisa.data.output.OutputHandler;
import monalisa.data.output.PetriNetOutputHandlers;


/**
 * Graphical Viewer for Petri nets and visualization of analysis results
 * @author Jens Einloft
 */
public class NetViewer extends JFrame implements ActionListener {
    private static final long serialVersionUID = 1113813463292202573L;

    // Constants declaration
    private static final Dimension DEFAULT_DIMENSION = new Dimension(1024, 768);
    private static final Dimension DOUBLE_DIMENSION = new Dimension(1024*2, 768*2);
    private static final Dimension TRIPPLE_DIMENSION = new Dimension(1024*3, 768*3);
    public static final String TRANSITION = "TANSITION";
    public static final String PLACE = "PLACE";
    public static final String BEND = "BEND";
    private static final String EXIT = "EXIT";
    private static final String HIDE_COLOR = "HIDE_COLOR";
    private static final String MAKE_PIC = "MAKE_PIC";
    private static final String SAVE_VIEW = "SAVE_LAYOUT";
    private static final String CTI_TRANSITIONS = "CTI_TRANSITIONS";
    private static final String BY_OCCURRENCE = "BY_OCCURRENCE";
    private static final String BY_FACTOR = "BY_FACTOR";
    private static final String SHOW_LABELS = "SHOW_LABELS";
    private static final String CENTER_NET = "CENTER_NET";
    private static final String RESET_COLORING = "RESET_COLORING";
    private static final String SHOW_COLOR_OPTION = "SHOW_COLOR_OPTION";

    private static final String NEW_VIEW = "NEW_VIEW";
    private static final String NODE_DEGREE  = "NODE_DEGREE";
    
    private static final Color defaultPanelColor = new Color(238,238,238);

    private static final ResourceManager resources = ResourceManager.instance();
    private static final StringResources strings = resources.getDefaultStrings();
    
    private Project project;
    private MainDialog mainDialog;
    private PetriNet petriNet;
    private TInvariants tinvs;
    private PInvariants pinvs;
    private Map<Configuration, Result> mctsResults;
    private Integer newEdgesCounter = 1, bendCounter = 1;
    protected Graph<NetViewerNode, NetViewerEdge> g;
    protected PersistentLayout<NetViewerNode, NetViewerEdge> layout;
    protected VisualizationViewer<NetViewerNode, NetViewerEdge> vv;
    private MyModalGraphMouse gm;
    private Boolean mouseMode;                                                  // true = Picking ; false = transforming
    private Boolean heatMap;                                                    // true = show , false = dont show
    private Boolean netChanged = false, lastChanged = false;
    private List<NetViewerNode> alignmentList = new ArrayList<NetViewerNode>(); // List to save selected nodes for an layout alignment

    private int viewCounter;
    private int vertexCounter;

    private MyUndoManager udm;
    private GraphPopupMousePlugin gpmp;
    private MyMouseListener mml;
    private NetViewerKeyListener nvkl;

    private VertexPaintTransformer vfpt = new VertexPaintTransformer();
    private VertexLabelTransformer vlt = new VertexLabelTransformer();
    private EdgePaintTransformer edpt;
    private EdgeLabelTransformer elt = new EdgeLabelTransformer();
    private int vertexSize, fontSize, edgeFactor;
    private float arrowSize;

    private Map<Integer, NetViewerNode> transitionMap = new HashMap<Integer, NetViewerNode>();
    private Map<Integer, NetViewerNode> placeMap = new HashMap<Integer, NetViewerNode>();
    protected Map<String, NetViewerNode> name2Node = new HashMap<String, NetViewerNode>();
    protected List<NetViewerNode> selection4Modification = new ArrayList<NetViewerNode>();
    private int placeCount, transitionCount, edgeCount;

    // Color Settings
    private Color tinvColor, mctsColor, heatMapColor, knockedOutColor, alsoKnockedOutColor, notKnockedOutColor, backgroundColor;
    private Map<String, Color> colorMap;

    // GUI Section
    private JMenuBar menuBar;
    private JMenu fileMenu, visualizationMenu, importantTransitonsMenu, viewMenu, topolocigalMenu, optionsMenu;
    private JMenuItem saveViewItem, newViewItem, exitItem, ctiItem, byOccurrenceItem, byFactorItem, centerItem,
                      resetColoringItem, colorOptionsItem, colorItem, labelItem, saveAsPictureItem, nodeDegreeItem;
    private JLabel messageLabel, infoBarLabel;
    private DefaultListModel neighborsListModel, allPlacesModel, allTransitionsModel, sisListModel;

    JSplitPane mainSplitPane;
    
    private ToolBarFrame tbf;
    private SearchBar sb;
    private VertexSetupFrame vsf;
    private EdgeSetupFrame esf;
    private LogicalPlacesFrame lpf;
    private ViewsPopupItemListener vpil;
    private ColorOptionsFrame cof;

    private List<JPanel> mouseMeodePanels;
    private List<JComboBox> tinvCbList;
    private List<JFrame> framesList;
    
    private Boolean viewsItemListerEnabeld = false;
    
    /**
     *
     * @param owner
     * @param project
     */
    public NetViewer(MainDialog owner, Project project) throws IOException, ClassNotFoundException, InterruptedException {
        this.project = project;
        this.petriNet = this.project.getPetriNet();
        this.mainDialog = owner;
        setLocationRelativeTo(owner);

        this.framesList = new ArrayList<JFrame>();
        this.tinvCbList = new ArrayList<JComboBox>();
        this.mouseMeodePanels = new ArrayList<JPanel>();               
        
        // Set default vertex and font size
        this.vertexSize = 12;
        this.fontSize = 12;
        this.arrowSize = 1;
        this.edgeFactor = 1;

        // Disable heatmaps for T-invariants
        this.heatMap = false;

        // Read the saved settings for the NetViewer out of config file
        tinvColor = Settings.getAsColor("tinvColor");
        mctsColor = Settings.getAsColor("mctsColor");
        heatMapColor = Settings.getAsColor("heatMapColor");
        knockedOutColor = Settings.getAsColor("knockedOutColor");
        alsoKnockedOutColor = Settings.getAsColor("alsoKnockedOutColor");
        notKnockedOutColor = Settings.getAsColor("notKnockedOutColor");
        backgroundColor = Settings.getAsColor("backgroundColor");

        colorMap = new HashMap<String, Color>();
        colorMap.put("tinvColor", tinvColor);
        colorMap.put("mctsColor", mctsColor);
        colorMap.put("heatMapColor", heatMapColor);
        colorMap.put("knockedOutColor", knockedOutColor);
        colorMap.put("alsoKnockedOutColor", alsoKnockedOutColor);
        colorMap.put("notKnockedOutColor", notKnockedOutColor);
        colorMap.put("backgroundColor", backgroundColor);
                
        initComponent();
    }

    /**
     * Init the NetViewer GUI and data structures
     */
    private void initComponent() throws ClassNotFoundException, InterruptedException {
        // --- Init Frame ---
        Dimension screenDimension = Toolkit.getDefaultToolkit().getScreenSize();  
        Dimension nvDimension = new Dimension((int)(screenDimension.getWidth()*0.95), (int)(screenDimension.getHeight()*0.95));
        
        setSize(nvDimension);
        setMinimumSize(new Dimension(800, 600));
        setTitle(strings.get("NVTitle")+" - "+project.getName());
        setIconImage(resources.getImage("icon-16.png"));
        setLocation(0, 0);

        addComponentListener(new ComponentListener() {
            
            Dimension currentDimension;
            double sizeOfRightSide,newSplitRatio;
            
            public void componentResized(ComponentEvent e) {
                currentDimension = ((JFrame)e.getSource()).getSize();
                sizeOfRightSide = currentDimension.getWidth() * 0.75;

                if(sizeOfRightSide > 400.0) {
                    newSplitRatio = 1 - (400 / currentDimension.getWidth());
                } else {
                    newSplitRatio = 0.75;                   
                }                
                
                mainSplitPane.setDividerLocation(newSplitRatio);
                mainSplitPane.setResizeWeight(newSplitRatio);     
                mainSplitPane.validate();                                
            }
            public void componentMoved(ComponentEvent e) { }
            public void componentShown(ComponentEvent e) { }
            public void componentHidden(ComponentEvent e) { }            
        });           
        
        // --- Graph Section ---
        g = new SparseGraph<NetViewerNode, NetViewerEdge>();
        edgeCount = 0;
        placeCount = 0;
        transitionCount = 0;
        NetViewerNode nvNode;
        int latesedId = 0;
        // generate Places
        String labelName, propertyName;
        for(Place place : this.petriNet.places()) {
            labelName = (String)place.getProperty("name");
            nvNode = new NetViewerNode(place.id(), NetViewer.PLACE, labelName);
            name2Node.put(labelName, nvNode);
            g.addVertex(nvNode);
            placeMap.put(place.id(), nvNode); // this map is needed to synchrone the PN with ne Netviewer graph and for coloring of the nodes
            placeCount++;            
            // save all properties from place to nvNode
            Iterator it = place.getPropertyList().iterator();            
            while(it.hasNext()) {                
                propertyName = (String) it.next();
                nvNode.putProperty(propertyName, place.getProperty(propertyName));
            }                  
            // serach for highest id, to get no conflict by creating new nodes
            if(place.id() > latesedId)
                latesedId = place.id();
        }

        // generate Transitions
        for(Transition transition : this.petriNet.transitions()) {
            labelName = (String)transition.getProperty("name");
            nvNode = new NetViewerNode(transition.id(), NetViewer.TRANSITION, labelName);
            name2Node.put(labelName, nvNode);
            g.addVertex(nvNode);
            transitionMap.put(transition.id(), nvNode);
            transitionCount++;
            // save all properties from place to nvNode
            Iterator it = transition.getPropertyList().iterator();            
            while(it.hasNext()) {
                propertyName = (String) it.next();
                nvNode.putProperty(propertyName, transition.getProperty(propertyName));
            }  
            // serach for highest id, to get no conflict by creating new nodes
            if(transition.id() > latesedId)
                latesedId = transition.id();
        }

        // generate Edges
        int weight;
        NetViewerEdge edge;
        for(Place place : this.petriNet.places()) {
            weight = 0;
            edge = null;
            // inputs
            for(Transition transition : place.inputs()) {
                weight = this.petriNet.getArc(petriNet.findTransition(transition.id()), petriNet.findPlace(place.id())).weight();
                edge = new NetViewerEdge("e"+edgeCount, weight, transitionMap.get(transition.id()), placeMap.get(place.id()));
                g.addEdge(edge, edge.getSource(), edge.getAim(), EdgeType.DIRECTED);
                edgeCount++;
            }
            // outputs
            for(Transition transition : place.outputs()) {
                weight = this.petriNet.getArc(petriNet.findPlace(place.id()), petriNet.findTransition(transition.id())).weight();
                edge = new NetViewerEdge("e"+edgeCount, weight, placeMap.get(place.id()), transitionMap.get(transition.id()));
                g.addEdge(edge, edge.getSource(), edge.getAim(), EdgeType.DIRECTED);
                edgeCount++;
            }
        }        

        vertexCounter = latesedId; // variable for the ids of new nodes

        // Set the graph layout
        layout = new PersistentLayoutImpl(new FRLayout(g));

        // Set size, depend on Petri net size
        if(transitionCount > 100 || placeCount > 100)
            layout.setSize(DOUBLE_DIMENSION);
        else if(transitionCount > 350 || placeCount > 350)
            layout.setSize(TRIPPLE_DIMENSION);
        else
            layout.setSize(DEFAULT_DIMENSION);

        // Init the VisualizationViewer and set all Renderer ect.
        vv = new VisualizationViewer<NetViewerNode, NetViewerEdge>(layout);//, gridSize, oo);
        vv.setPreferredSize(nvDimension);
        vv.setSize(nvDimension);
        vv.setVertexToolTipTransformer(new VertexToolTipTransformer()); // render the tooltips by mouse over
        vv.setEdgeToolTipTransformer(new EdgeToolTipTransformer());
        vv.getRenderer().setEdgeRenderer(new MyEdgeRenderer());
        vv.getRenderContext().setVertexLabelTransformer(vlt); // render vertex label
        vv.getRenderContext().setEdgeLabelTransformer(elt); // render edge label
        vv.getRenderContext().setVertexShapeTransformer(new VertexShapeTransformer(vertexSize)); // render the shape of the vertices
        vv.getRenderContext().setVertexFillPaintTransformer(vfpt); // controll the color of vertices
        vv.getRenderContext().setEdgeShapeTransformer(new EdgeShape.Line()); // render the border of vertices, because bends need no border
        vv.getRenderContext().setEdgeStrokeTransformer(new EdgeStrokeTransformer(edgeFactor));
        vv.getRenderContext().setVertexFontTransformer(new ConstantTransformer(new Font("Helvetica", Font.BOLD, fontSize)));
        vv.getRenderContext().setEdgeFontTransformer(new ConstantTransformer(new Font("Helvetica", Font.BOLD, fontSize)));
        vv.getRenderContext().setEdgeIncludePredicate(new EdgeIncludePredicate()); // controll the visibility of edges
        vv.getRenderContext().setVertexDrawPaintTransformer(new VertexDrawPaintTransformer()); // controll the color of the edges
        edpt = new EdgePaintTransformer(vv); // Edge + arrow fill and border color
        vv.getRenderContext().setEdgeDrawPaintTransformer(edpt);
        vv.getRenderContext().setArrowDrawPaintTransformer(edpt);
        vv.getRenderContext().setArrowFillPaintTransformer(edpt);
        vv.getRenderContext().setEdgeArrowTransformer(new EdgeArrowTransformer(arrowSize)); // control the size of the arrows
        vv.getRenderer().setVertexLabelRenderer(new AdvancedVertexLabelRenderer());

        // Create the message label at the top of die vv
        messageLabel = new JLabel();
        vv.add(messageLabel);

        // Init all MouseListener and MousePlugins        
        gpmp = new GraphPopupMousePlugin(this);
        gm = new MyModalGraphMouse(gpmp, this); // Mouse for picking and transforming the graph        
        gm.add(gpmp);
        vv.setGraphMouse(gm);
        mouseMode = true;
                
        mml = new MyMouseListener(this); // Mouselistener for creating new places and transitions to model a Petri net
        vv.addMouseListener(mml);

        nvkl = new NetViewerKeyListener(this); // KeyListener for hotkeys
        vv.addKeyListener(nvkl);
        addKeyListener(nvkl);

        // If an vertex is selected, the MousePlugin must called for possible actions
        vv.getRenderContext().getPickedVertexState().addItemListener(new ItemListener() {
            @Override
            public void itemStateChanged(ItemEvent e) {
                if(e.getStateChange() == ItemEvent.SELECTED) {
                    gpmp.handlePicking((NetViewerNode)e.getItem());
                }
            }
        });

        // If an edge is selected, the MousePlugin must called for possible actions
        vv.getRenderContext().getPickedEdgeState().addItemListener(new ItemListener() {
            @Override
            public void itemStateChanged(ItemEvent e) {
                if(e.getStateChange() == ItemEvent.SELECTED) {
                    gpmp.handlePicking((NetViewerEdge)e.getItem());
                }
            }
        });

        vv.setBackground(backgroundColor);

        // START --- GUI building section

        // Create Frames like ToolBar etc and connect them with NetViewer
        initExternalFrames();

        // Init MenuBar
        initMenuBar();

        infoBarLabel = new JLabel();
        infoBarLabel.setSize(new Dimension((int) nvDimension.getWidth(), 20));
        infoBarLabel.setPreferredSize(new Dimension((int) nvDimension.getWidth(), 20));

        mainSplitPane = new JSplitPane();
        mainSplitPane.setSize(nvDimension);
        mainSplitPane.setPreferredSize(nvDimension);
        mainSplitPane.setOneTouchExpandable(true);
        mainSplitPane.setVerifyInputWhenFocusTarget(false);
        
        mainSplitPane.setLeftComponent(vv);
        mainSplitPane.setRightComponent(tbf.getContentPane());        
        
        Container contentPane = getContentPane();
        double sizeOfMainLayout[][] =
        {{TableLayout.FILL},
        {TableLayout.FILL, TableLayout.PREFERRED}};

        contentPane.setLayout(new TableLayout(sizeOfMainLayout));
        contentPane.setSize(nvDimension);
        contentPane.add(mainSplitPane, "0,0");
        contentPane.add(infoBarLabel, "0,1");
        
        pack();
        
        updateInfoBar();        
        
        

        updateSearchBar(g.getVertices());
        markSelectedMouseMode(tbf.mousePickingPanel);

        // END --- GUI building section
        
        // Relocate vertices, if information about their positions are available from the Petri net
        for(Place place : this.petriNet.places()) {
            if(place.hasProperty("posX")) {
                nvNode = placeMap.get(place.id());
                layout.setLocation(nvNode, new Point2D.Double((Double)place.getProperty("posX"), (Double)place.getProperty("posY")));
            }
        }

        for(Transition transition : this.petriNet.transitions()) {
            if(transition.hasProperty("posX")) {
                nvNode = transitionMap.get(transition.id());
                layout.setLocation(nvNode, new Point2D.Double((Double)transition.getProperty("posX"), (Double)transition.getProperty("posY")));
            }
        }
        vv.repaint();

        // Add all Views to the ComboBox
        viewCounter = project.getViewWrapper().getViewCounter();
        if(viewCounter > 0) {
            for(View v : project.getViewWrapper().getViews())
                tbf.viewsCb.addItem(v);            
        }
        updateViewsCb();
        
        // Select the latest view
        viewsItemListerEnabeld = true;
        if(viewCounter > 0) {
            tbf.viewsCb.setSelectedIndex(project.getViewWrapper().getViewCounter()-1);  
            
            if(viewCounter == 1) {                
                vpil.itemStateChanged(new ItemEvent(tbf.viewsCb, ItemEvent.ITEM_STATE_CHANGED, tbf.viewsCb.getSelectedItem(), ItemEvent.SELECTED));
            }            
        }

        netChanged = false;
        netChanged();

        // Init UndoManager
        udm = new MyUndoManager();        
    }

    /**
     * Which action is to perform?
     * @param e
     */
    @Override
    public void actionPerformed(ActionEvent e) {
        final String action = e.getActionCommand();

        if(action == HIDE_COLOR)
            hideColor();
        else if(action == MAKE_PIC) {
            try {
                makePic();
            } catch (IOException ex) {
                Logger.getLogger(NetViewer.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        else if(action == NEW_VIEW)
            try {
                saveView(false, false);
            } catch (ClassNotFoundException ex) {
                Logger.getLogger(NetViewer.class.getName()).log(Level.SEVERE, null, ex);
            } catch (IOException ex) {
                Logger.getLogger(NetViewer.class.getName()).log(Level.SEVERE, null, ex);
            }
        else if(action == SAVE_VIEW)
            try {
                saveView(false, true);
            } catch (ClassNotFoundException ex) {
                Logger.getLogger(NetViewer.class.getName()).log(Level.SEVERE, null, ex);
            } catch (IOException ex) {
                Logger.getLogger(NetViewer.class.getName()).log(Level.SEVERE, null, ex);
            }
        else if(action == CENTER_NET)
            center();
        else if(action  == CTI_TRANSITIONS)
            colorCTITransitions();
        else if(action == BY_OCCURRENCE)
            colorImportantTransitions(true);
        else if(action == BY_FACTOR)
            colorImportantTransitions(false);
        else if(action == SHOW_LABELS)
            showLabels();
        else if(action == SHOW_COLOR_OPTION)
            showColorOptions();
        else if(action == NODE_DEGREE)
            showNodeDegrees();
        else if(action == RESET_COLORING)
            resetColor();        
        else if(action == EXIT)
            exitNetViewer();
    }

    private int getNewNodeId() {
        return ++vertexCounter;
    }

    /**
     * Centers the Petri net in the VisualizationViewer
     */
    protected void center() {
        vv.getRenderContext().getMultiLayerTransformer().getTransformer(Layer.LAYOUT).setToIdentity();
        vv.getRenderContext().getMultiLayerTransformer().getTransformer(Layer.VIEW).setToIdentity();
        setZoomScale(100.0);
    }

    protected void undoLastAction() {
        udm.undo();
    }

    /**
     * If a node is create in the VisualizationViewer, this function will update the Petri net
     * @param nvNode
     */
    private void addVertexToPetriNet(NetViewerNode nvNode) {
        Point2D point = layout.transform(nvNode);
        if(nvNode.getNodeType().equalsIgnoreCase(NetViewer.PLACE)) {
            Place place = new Place(nvNode.getId());
            place.putProperty("name",nvNode.getName());
            place.putProperty("posX", point.getX());
            place.putProperty("posY", point.getY());
            project.getPetriNet().addPlace(place);
            placeMap.put(place.id(), nvNode);
        }
        else if(nvNode.getNodeType().equalsIgnoreCase(NetViewer.TRANSITION)) {
            Transition transition = new Transition(nvNode.getId());
            transition.putProperty("name", nvNode.getName());
            transition.putProperty("posX", point.getX());
            transition.putProperty("posY", point.getY());
            project.getPetriNet().addTransition(transition);
            transitionMap.put(transition.id(), nvNode);
        }
    }

    /**
     * If a node is removed in the VisualizationViewer, this function will update the Petri net
     * @param nvNode
     */
    private void removeVertexFromPetriNet(NetViewerNode nvNode) {
         if(nvNode.getNodeType().equalsIgnoreCase(NetViewer.PLACE)) {
            project.getPetriNet().removePlace(project.getPetriNet().findPlace(nvNode.getId()));
            placeMap.remove(nvNode.getId());
        }
        else if(nvNode.getNodeType().equalsIgnoreCase(NetViewer.TRANSITION)) {
            project.getPetriNet().removeTransition(project.getPetriNet().findTransition(nvNode.getId()));
            transitionMap.remove(nvNode.getId());
        }
        nvNode = null;
    }

    /**
     * If a edge is create in the VisualizationViewer, this function will update the Petri net
     * @param edge
     */
    private void addArcToPetriNet(NetViewerEdge edge) {
        Object from, to;
        Arc arc = null;
        if(edge.getSource().getNodeType().equals(NetViewer.PLACE)) {
            from = project.getPetriNet().findPlace(edge.getSource().getMasterNode().getId());
            to = project.getPetriNet().findTransition(edge.getAim().getId());
            arc = new Arc(from, to, edge.getWeight());
            project.getPetriNet().addArc((Place)from, (Transition)to, arc);
        }
        else if(edge.getSource().getNodeType().equals(NetViewer.TRANSITION)) {
            from = project.getPetriNet().findTransition(edge.getSource().getId());
            to = project.getPetriNet().findPlace(edge.getAim().getMasterNode().getId());
            arc =  new Arc(from, to, edge.getWeight());
            project.getPetriNet().addArc((Transition)from, (Place)to, arc);
        }
    }

    /**
     * If a edge is removed in the VisualizationViewer, this function will update the Petri net
     * @param edge
     */
    private void removeArcFromPetriNet(NetViewerEdge edge) {
        if(edge.getSource().getNodeType().equals(NetViewer.PLACE)) {
            Place from = project.getPetriNet().findPlace(edge.getSource().getMasterNode().getId());
            Transition to = project.getPetriNet().findTransition(edge.getAim().getId());
            Arc arc = project.getPetriNet().getArc(from, to);
            if(arc != null)
                project.getPetriNet().removeArc(from, to);
        }
        else {
            Transition from = project.getPetriNet().findTransition(edge.getSource().getId());
            Place to = project.getPetriNet().findPlace(edge.getAim().getMasterNode().getId());
            Arc arc = project.getPetriNet().getArc(from, to);
            if(arc != null)
                project.getPetriNet().removeArc(from, to);
        }
        edge = null;
    }

    /**
     * Creates a new node in the VisualizationViewer at position (X,Y)
     * @param type
     * @param x
     * @param y
     */
    protected void createNewNode(String type, int x, int y) {
        NetViewerNode newNode = null;
        String name;
        if(type.equalsIgnoreCase(NetViewer.PLACE)) {
            name = "P"+(++placeCount);
            createNewNode(NetViewer.PLACE, name, x, y, false);
        }
        else if(type.equalsIgnoreCase(NetViewer.TRANSITION)) {
            name = "T"+(++transitionCount);
            createNewNode(NetViewer.TRANSITION, name, x, y, false);
        }
    }
    
    /**
     * Creates a new node in the VisualizationViewer at position (X,Y) with given labelName
     * @param type
     * @param labelName
     * @param x
     * @param y
     */
    protected NetViewerNode createNewNode(String type, String labelName, int x, int y, Boolean undo) {
        int id = getNewNodeId();
        NetViewerNode newNode = new NetViewerNode(id, type, labelName);
        
        g.addVertex(newNode);
        addVertexToPetriNet(newNode);

        if(!undo)
            udm.addEvent(new UndoEvent(UndoEvent.ADD_VERTEX, newNode));

        Point point = new Point(x,y);
       
        layout.setLocation(newNode, vv.getRenderContext().getMultiLayerTransformer().inverseTransform(point));

        netChanged = true;
        netChanged();   

        updateSearchBar(g.getVertices());
        updateInfoBar();
        vv.repaint();
        project.setProjectChanged(true);
        
        return newNode;
    }

    /**
     * @param nvNode
     * @param direction 0 = input ; 1 = output
     * @param x
     * @param y
     */
    protected void createNewNode(NetViewerNode nvNode, Boolean direction) {
        NetViewerEdge newEdge;       
        NetViewerNode newNode = null;

        int vertexNumber = getNewNodeId();

        if((nvNode.getNodeType().equalsIgnoreCase(NetViewer.TRANSITION)))
            newNode = new NetViewerNode(vertexNumber, NetViewer.PLACE, "P"+(++placeCount));
        else
            newNode = new NetViewerNode(vertexNumber, NetViewer.TRANSITION, "T"+(++transitionCount));
        g.addVertex(newNode);
        addVertexToPetriNet(newNode);

        if(direction)
            newEdge = new NetViewerEdge("n"+newEdgesCounter+(int)(Math.random() * (1000 - 1) + 1), 1, newNode, nvNode);
        else
            newEdge = new NetViewerEdge("n"+newEdgesCounter+(int)(Math.random() * (1000 - 1) + 1), 1, nvNode, newNode);
        
        g.addEdge(newEdge, newEdge.getSource(), newEdge.getAim(), EdgeType.DIRECTED);
        addArcToPetriNet(newEdge);
        newEdgesCounter++;

        udm.addEvent(new UndoEvent(UndoEvent.ADD_VERTEX, newNode));

        netChanged = true;
        netChanged();        

        layout.setLocation(newNode, calcNewPoint(layout.transform(nvNode)));
        
        updateSearchBar(g.getVertices());
        updateInfoBar();
        vv.repaint();
        project.setProjectChanged(true);
    }

    /**
     * Create a reversed transition of a given transition
     * @param nvNode
     * @param x
     * @param y
     */
    protected void createReverseTransition(NetViewerNode nvNode, int x, int y) {
        if(!nvNode.getNodeType().equalsIgnoreCase(NetViewer.TRANSITION))
            return;
        
        NetViewerNode newNode = createNewNode(TRANSITION, nvNode.getName()+"_rev" , x, y, false);

        for(NetViewerEdge e : nvNode.getInEdges())
            createNewEdge(newNode, e.getSource(), e.getWeight(), false);

        for(NetViewerEdge e : nvNode.getOutEdges())
            createNewEdge(e.getAim(), newNode, e.getWeight(), false);

        udm.addEvent(new UndoEvent(UndoEvent.REVERSE_TRANSITION, newNode));
    }

    /**
     * Deletes a node from the VisualizationViewer
     * @param nvNode
     */
    protected void deleteNode(NetViewerNode nvNode, Boolean undo) {
        if(nvNode == null)
            return;
        if(nvNode.isMasterNode()) {
            List<NetViewerNode> nvList = new ArrayList<NetViewerNode>();
            for(NetViewerNode n : nvNode.getLogicalPlaces())
                nvList.add(n);
            for(NetViewerNode n : nvList)
                removeLogicalPlace(n, null, false);
            nvNode.getLogicalPlaces().clear();
        }
        else if(nvNode.isLogical()) {
            this.removeLogicalPlace(nvNode, null, false);
            return;
        }
        
        // Remove all bends
        List<NetViewerEdge> bends2remove = new ArrayList<NetViewerEdge>();
        for(NetViewerEdge edge : nvNode.getInEdges()) {                       
            if(!edge.getMasterEdge().equals(edge))
                continue;            
            if(!edge.getMasterEdge().getBendEdges().isEmpty()) {
                bends2remove.add(edge.getMasterEdge());
            }
        }        
        for(NetViewerEdge edge : nvNode.getOutEdges()) {
            if(!edge.getMasterEdge().equals(edge))
                continue;            
            if(!edge.getMasterEdge().getBendEdges().isEmpty()) {
                bends2remove.add(edge.getMasterEdge());
            }         
        }        
        for(NetViewerEdge edge : bends2remove) {
            planeEdge(edge, false);
        }
        
        // Now remove all edges
        for(NetViewerEdge edge : nvNode.getInEdges()) {
            removeArcFromPetriNet(edge.getMasterEdge());
            g.removeEdge(edge.getMasterEdge());
            edge.getSource().removeOutEdge(edge);
            edge = null;
        }

        for(NetViewerEdge edge : nvNode.getOutEdges()) {
            removeArcFromPetriNet(edge.getMasterEdge());
            g.removeEdge(edge.getMasterEdge());            
            edge.getAim().removeInEdge(edge.getMasterEdge());
            edge = null;
        }
        
        // and than the node
        g.removeVertex(nvNode);
        if(!nvNode.isLogical())
            removeVertexFromPetriNet(nvNode);

        netChanged = true;
        netChanged();
        
        if(!undo)
            udm.addEvent(new UndoEvent(UndoEvent.DELETE_VERTEX, nvNode, layout.transform(nvNode)));

        vv.getPickedVertexState().clear();
        
        updateSearchBar(g.getVertices());
        updateInfoBar();
        vv.repaint();
        project.setProjectChanged(true);
    }


    /**
     * Removes all bends from an edge
     * @param nvEdge
     * @param repaint
     */
    protected void planeEdge(NetViewerEdge nvEdge, Boolean repaint) {
        removeBendEdges(nvEdge);
        if(repaint)
            vv.repaint();
    }

    /**
     * Creates a new edge between aim and source
     * @param source
     * @param aim
     * @return true if a new edge is created
     */
    protected Boolean createNewEdge(NetViewerNode source, NetViewerNode aim) {
        return createNewEdge(source, aim, 1, false);
    }

    /**
    * Creates a new edge between aim and source with a given weight
    * @param source
    * @param aim
    * @param weight
    * @return true if a new edge is created
    */
    protected Boolean createNewEdge(NetViewerNode source, NetViewerNode aim, int weight, Boolean undo) {
        Arc arc = null;
        if(source.getNodeType().equals(NetViewer.PLACE)) {
            arc = project.getPetriNet().getArc(project.getPetriNet().findPlace(source.getMasterNode().getId()), project.getPetriNet().findTransition(aim.getMasterNode().getId()));
            if(arc != null)
                return false;
        }
        else if(source.getNodeType().equals(NetViewer.TRANSITION) && (aim.isLogical() || aim.isMasterNode())) {
            arc = project.getPetriNet().getArc(project.getPetriNet().findTransition(source.getMasterNode().getId()), project.getPetriNet().findPlace(aim.getMasterNode().getId()));
            if(arc != null)
                return false;
        }

        if(g.findEdge(source, aim) == null) {
            NetViewerEdge newEdge = new NetViewerEdge("n"+newEdgesCounter+(int)(Math.random() * (1000 - 1) + 1), weight, source, aim);
            newEdgesCounter++;

            g.addEdge(newEdge, source, aim, EdgeType.DIRECTED);
            addArcToPetriNet(newEdge);

            netChanged = true;
            netChanged();

            if(!undo)
                udm.addEvent(new UndoEvent(UndoEvent.ADD_EDGE, newEdge));

            markSelectedMouseMode(tbf.mousePickingPanel);

            updateInfoBar();
            vv.repaint();
            project.setProjectChanged(true);
            
            return true;
        }
        
        return false;
    }

    /**
     * Delete an edge from the VisualizationViewer
     * @param nvEdge
     * @param undo
     */
    protected void deleteEdge(NetViewerEdge nvEdge, Boolean undo) {
        if(nvEdge == null)
            return;
        if(nvEdge.getMasterEdge().equals(nvEdge)) {
            nvEdge.getAim().removeInEdge(nvEdge);
            nvEdge.getSource().removeOutEdge(nvEdge);
        }

        removeBendEdges(nvEdge);

        nvEdge.getMasterEdge().getAim().removeInEdge(nvEdge.getMasterEdge());
        nvEdge.getMasterEdge().getSource().removeOutEdge(nvEdge.getMasterEdge());

        g.removeEdge(nvEdge.getMasterEdge());

        removeArcFromPetriNet(nvEdge.getMasterEdge());

        netChanged = true;
        netChanged();

        if(!undo)
            udm.addEvent(new UndoEvent(UndoEvent.DELETE_EDGE, nvEdge));

        updateInfoBar();
        vv.getPickedEdgeState().clear();
        vv.repaint();
        project.setProjectChanged(true);
    }

    /**
     * Save the current color settings
     * @throws FileNotFoundException
     * @throws IOException
     */
    protected void saveColorOptions() throws FileNotFoundException, IOException {
        Settings.setColorOption("tinvColor", cof.invColorSelectionLabel.getBackground());
        Settings.setColorOption("mctsColor", cof.mctsColorSelectionLabel.getBackground());
        Settings.setColorOption("heatMapColor", cof.heatMapColorSelectionLabel.getBackground());
        Settings.setColorOption("knockedOutColor", cof.knockedOutColorSelectionLabel.getBackground());
        Settings.setColorOption("alsoKnockedOutColor", cof.alsoKnockedOutColorSelectionLabel.getBackground());
        Settings.setColorOption("notKnockedOutColor", cof.notKnockedOutColorSelectionLabel.getBackground());
        Settings.setColorOption("backgroundColor", cof.backgroundColorSelectionLabel.getBackground());

        Settings.writeToFile(System.getProperty("user.home")+"/.monalisaSettings");

        tinvColor = Settings.getAsColor("tinvColor");
        mctsColor = Settings.getAsColor("mctsColor");
        heatMapColor = Settings.getAsColor("heatMapColor");
        knockedOutColor = Settings.getAsColor("knockedOutColor");
        alsoKnockedOutColor = Settings.getAsColor("alsoKnockedOutColor");
        notKnockedOutColor = Settings.getAsColor("notKnockedOutColor");
        backgroundColor = Settings.getAsColor("backgroundColor");

        vv.setBackground(backgroundColor);
        vv.repaint();
        cof.setVisible(false);
        this.setEnabled(true);
    }

    /**
     *
     * @param value
     */
    public void setNetChanged(Boolean value) {
        netChanged = value;
    }

    /**
     * Checks if the Petri net is changed and disable / enable gui elemets
     */
    private void netChanged() {
        if(netChanged) {
            if(project.hasResults(new TInvariantTool()) || project.hasResults(new PInvariantTool()) || project.hasResults(new MctsTool())) {
                messageLabel.setText(strings.get("NVNetChanged"));
                messageLabel.setForeground(Color.RED);
            }
        }
        if(netChanged && !lastChanged) {
            lastChanged = netChanged;

            if(project.hasResults(new TInvariantTool()) || project.hasResults(new PInvariantTool()) || project.hasResults(new MctsTool())) {
                messageLabel.setText(strings.get("NVNetChanged"));
                messageLabel.setForeground(Color.RED);
                //tinvCb.setEnabled(false);
                //tbf.ioInvCb.setEnabled(false);
                //tbf.cyclicInvCb.setEnabled(false);
                //tbf.mctsCb.setEnabled(false);
                //importantTransitonsMenu.setEnabled(false);
                
                tinvs = null;
                for(JComboBox cb : tinvCbList)
                    cb.removeAllItems();
                mctsResults = null;
                tbf.mctsCb.removeAllItems();
                project.resetTools();
                mainDialog.updateUI();
            }
        }
        else if(!netChanged && lastChanged){
            lastChanged = netChanged;
            
            resetMessageLabel();
            if(hasTinvs()) {
                for(JComboBox cb : tinvCbList)
                    cb.setEnabled(true);
            }
            if(hasMcts())
                tbf.mctsCb.setEnabled(true);
            tbf.saveViewButton.setEnabled(true);
            saveViewItem.setEnabled(true);
            tbf.importViewButton.setEnabled(true);
            tbf.exportViewButton.setEnabled(true);
            importantTransitonsMenu.setEnabled(true);
        }
    }
    
    public Boolean hasNetChanged() {
        return netChanged;
    }

   /**
    * Sets the font size to a given value
    * @param newFontSize 
    */
    public void setFontSize(int newFontSize) {
        vv.getRenderContext().setVertexFontTransformer(new ConstantTransformer(new Font("Helvetica", Font.BOLD, newFontSize)));
        vv.getRenderContext().setEdgeFontTransformer(new ConstantTransformer(new Font("Helvetica", Font.BOLD, newFontSize)));
        vv.repaint();        
    }
    
    /**
     * Sets the icon size to a given value
     * @param newVertexSize 
     */
    public void setIconSize(int newVertexSize) {
        vv.getRenderContext().setVertexShapeTransformer(new VertexShapeTransformer(newVertexSize));
        vv.repaint();        
    }

    /**
     * Sets the arrow size to a given value
     * @param newArrowSize 
     */
    public void setArrowSize(double newArrowSize) {
        vv.getRenderContext().setEdgeArrowTransformer(new EdgeArrowTransformer(new Double(newArrowSize).floatValue()));
        vv.repaint();        
    }
    
    /**
     * Sets the edge size to a given value
     * @param newEdgeSize 
     */
    public void setEdgeSize(int newEdgeSize) {
        vv.getRenderContext().setEdgeStrokeTransformer(new EdgeStrokeTransformer(newEdgeSize));
        vv.repaint();        
    }    

    /**
     * Zooms in or out by a given value.
     * @param setSlider
     */
    protected void zommToValue(int inOrOut) {
        Point2D center = new Point(vv.getSize().height/2, vv.getSize().width/2);               
        
        if(inOrOut > 0) {
            this.gm.getScalingPlugin().getScaler().scale(vv, 1.1F, center);
        }
        else if(inOrOut < 0) {
            this.gm.getScalingPlugin().getScaler().scale(vv, 0.9090909F, center);                 
        }
        
        double viewScale = vv.getRenderContext().getMultiLayerTransformer().getTransformer(Layer.VIEW).getScale();
        double layoutScale = vv.getRenderContext().getMultiLayerTransformer().getTransformer(Layer.LAYOUT).getScale();     
        setZoomScale((viewScale*layoutScale)*100);             
    }    
    
    /**
     * Set the value of the JSPinner in the ToolBarFrame
     * @param zoomScale 
     */
    protected void setZoomScale(double zoomScale) {      
        tbf.setZoomSpinnerValue(new Double(zoomScale).intValue());     
    }
     
    protected Boolean heatMap() {
        if(heatMap) {
            heatMap = false;
            return heatMap;
        }
        else {
            heatMap = true;
            return heatMap;
        }
    }

    /**
     * true = picking, false = transforming
     * @return
     */
    protected Boolean getMouseMode() {
        return mouseMode;
    }

    /**
     * Change mousemode to transforming
     */
    protected void changeMouseModeToTransforming() {
        cancelMouseAction();
        mouseMode = false;
        markSelectedMouseMode(tbf.mouseTransformingPanel);
        gm.setMode(ModalGraphMouse.Mode.TRANSFORMING);
        netChanged();
    }

    /**
     * change mousemode to picking
     */
    protected void changeMouseModeToPicking() {
        cancelMouseAction();
        mouseMode = true;
        markSelectedMouseMode(tbf.mousePickingPanel);
        gm.setMode(ModalGraphMouse.Mode.PICKING);
        vv.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        netChanged();
    }

    /**
     * Returns all t-invariants form loaded petrinet
     * @return
     */
    private TInvariants getTInvs() {
        return project.getResult(TInvariantTool.class, new TInvariantsConfiguration());
    }

    /**
     * Returns all p-invariants form loaded Petri net
     * @return
     */
    private PInvariants getPInvs() {
        return project.getResult(PInvariantTool.class, new PInvariantsConfiguration());
    }

    /**
     * Return all MCT-sets from loaded Petri net
     * @return
     */
    private Map<Configuration, Result> getMctsResults() {
        return project.getResults(MctsTool.class);
    }

    /**
     * Hide coloring of vertex
     */
    protected Boolean hideColor() {
        vfpt.setHideColor((vfpt.getHideColor()) ? false : true);
        edpt.setHideColor((edpt.getHideColor()) ? false : true);

        if(!vfpt.getHideColor()) {
            colorItem.setText(strings.get("NVShowColorText"));
            vv.repaint();
            return false;
        }
        else {
            colorItem.setText(strings.get("NVHideColorText"));            
            vv.repaint();
            return true;
        }
    }

    protected Boolean showLabels() {
        if(vlt.showLabel()) {
            vlt.setShowLabel(false);
            labelItem.setText(strings.get("NVShowAllLabels"));

            vv.repaint();
            return false;
        }
        else {
            vlt.setShowLabel(true);
            labelItem.setText(strings.get("NVHideAllLabels"));

            vv.repaint();
            return true;
        }        
    }

    /**
     * Save the current layout in a PNG or SVG file
     * @throws IOException
     */
    public void makePic() throws IOException {
        messageLabel.setVisible(false);

        vv.getPickedVertexState().clear();
        vv.getPickedEdgeState().clear();

        MyFileFilter pngFilter = new MyFileFilter("png", "Portable Network Graphics");
        MyFileFilter svgFilter = new MyFileFilter("svg", "Support Vector Graphics");
        
        JFileChooser imgPathChooser = null;
        if(project != null) {
            if(project.getPath() != null)
                imgPathChooser = new JFileChooser(project.getPath().getPath());
            else
                imgPathChooser = new JFileChooser();
        }
        else {
            imgPathChooser = new JFileChooser();
        }
        
        imgPathChooser.setDialogTitle(strings.get("NVImageLocation"));
        imgPathChooser.setAcceptAllFileFilterUsed(false);
        imgPathChooser.addChoosableFileFilter(pngFilter);
        imgPathChooser.addChoosableFileFilter(svgFilter);
        imgPathChooser.showSaveDialog(this);

        File imgFile = imgPathChooser.getSelectedFile();
        if(imgFile != null) {
            MyFileFilter selectedFileFilter = ((MyFileFilter)imgPathChooser.getFileFilter());

            // Are an ".png" or ".svg" at the end of the filename?
            imgFile = selectedFileFilter.checkFileNameForExtension(imgFile);

            if(selectedFileFilter.getExtension().equalsIgnoreCase("png")) {                
                BufferedImage img = new BufferedImage(vv.getSize().width, vv.getSize().height, BufferedImage.TYPE_INT_RGB);
                vv.paintAll(img.getGraphics());
                ImageIO.write(img, "png", imgFile);
            }
            else if(selectedFileFilter.getExtension().equalsIgnoreCase("svg")) {
                // Get a DOMImplementation
                DOMImplementation domImpl = GenericDOMImplementation.getDOMImplementation();
                // Create an instance of org.w3c.dom.Document
                Document document = domImpl.createDocument("http://www.w3.org/2000/svg", "svg", null);
                // Create an instance of the SVG Generator
                SVGGraphics2D svgGenerator = new SVGGraphics2D(document);
                vv.paint(svgGenerator);
                svgGenerator.stream(imgFile.getAbsolutePath());
            }
        }        
        messageLabel.setVisible(true);
    }

    public void addTinvsToComboBox() {
        if(tinvs == null)
            tinvs = getTInvs();
        if(tinvs == null)
            return;
        addTinvsToComboBox(tinvs);
    }

    /**
     * Add the T-Invariants to the combo box
     */
    private void addTinvsToComboBox(TInvariants tinvs) {
        if(tinvs == null)
            return;
        if(!tinvs.isEmpty()) {
            for(JComboBox cb : tinvCbList) {
                cb.setEnabled(true);
                cb.removeAllItems();
            }

            Boolean input, output, row, col, plus, minus;
            int adjMatrix[][] = null, i = 0, j = 0, tinvSize;

            Iterator<Transition> it1,it2;
            Transition t1, t2;
            for(TInvariant tinv : tinvs) {
                tbf.allInvCb.addItem(new TinvWrapper(tinv));

                tinvSize = tinv.size();
                if(tinvSize == 2) {
                    tbf.trivialInvCb.addItem(new TinvWrapper(tinv));
                    continue;
                }


                adjMatrix = new int[tinvSize][tinvSize];
                i = 0;
   
                input = false;
                output = false;
                row = false;
                col = false;

                it1 = tinv.transitions().iterator();
                while(it1.hasNext()) {
                    j = 0;
                    t1 = it1.next();

                    // Check for I/O TInvariant
                    if(transitionMap.get(t1.id()).getInEdges().isEmpty())
                        input = true;
                    if(transitionMap.get(t1.id()).getOutEdges().isEmpty())
                        output = true;
                }

                if(!output && !input)
                    tbf.cyclicInvCb.addItem(new TinvWrapper(tinv));
                else if(output && input)
                    tbf.ioInvCb.addItem(new TinvWrapper(tinv));
                else if(output && !input)
                    tbf.outputInvCb.addItem(new TinvWrapper(tinv));
                else if(input && !output)
                    tbf.inputInvCb.addItem(new TinvWrapper(tinv));
            }

            // Create a T-invariant out of all t-invariantzs of a single combobox
            int itemCount, oldValue;
            TInvariant tinvariant;
            Map<Transition, Integer> transitions = new HashMap<Transition, Integer>();
            for(JComboBox cb : tinvCbList) {
                itemCount = cb.getItemCount();
                if(itemCount > 0) {
                    tinvariant = null;
                    transitions = new HashMap<Transition, Integer>();
                    for(i = 0; i < itemCount; i++) {
                        tinvariant = ((TinvWrapper)cb.getItemAt(i)).getTinv();
                        for(Transition transition : tinvariant) {
                            if(tinvariant.factor(transition) > 0) {
                                if(!transitions.containsKey(transition))
                                    transitions.put(transition, 0);
                                oldValue = transitions.get(transition);
                                transitions.put(transition, tinvariant.factor(transition)+oldValue);
                            }
                        }
                    }
                    cb.insertItemAt(new TinvWrapper(new TInvariant(-1, transitions)), 0);
                }
            }

            tbf.allInvCb.insertItemAt(strings.get("NVAllT", (tbf.allInvCb.getItemCount() == 0) ? 0 : tbf.allInvCb.getItemCount()-1), 0);
            tbf.ioInvCb.insertItemAt(strings.get("NVIOT", (tbf.ioInvCb.getItemCount() == 0) ? 0 : tbf.ioInvCb.getItemCount()-1), 0);
            tbf.trivialInvCb.insertItemAt(strings.get("NVTrivialT", (tbf.trivialInvCb.getItemCount() == 0) ? 0 : tbf.trivialInvCb.getItemCount()-1), 0);
            tbf.outputInvCb.insertItemAt(strings.get("NVOOT", (tbf.outputInvCb.getItemCount() == 0) ? 0 : tbf.outputInvCb.getItemCount()-1), 0);
            tbf.inputInvCb.insertItemAt(strings.get("NVOIT", (tbf.inputInvCb.getItemCount() == 0) ? 0 : tbf.inputInvCb.getItemCount()-1), 0);
            tbf.cyclicInvCb.insertItemAt(strings.get("NVCyclicT", (tbf.cyclicInvCb.getItemCount() == 0) ? 0 : tbf.cyclicInvCb.getItemCount()-1), 0);

            for(JComboBox cb : tinvCbList)
                cb.setSelectedIndex(0);
        }
    }

    /**
     * Add the T-Invariants to the combobox
     */
    public void addPinvsToComboBox() {
        if(pinvs == null)
            pinvs = getPInvs();
        if(pinvs == null)
            return;
        if(!pinvs.isEmpty()) {
            tbf.pinvCb.setEnabled(true);
            tbf.pinvCb.removeAllItems();

            for(PInvariant pinv : pinvs)
                tbf.pinvCb.addItem(new PinvWrapper(pinv));

            tbf.pinvCb.insertItemAt(strings.get("NVAllP", tbf.pinvCb.getItemCount()), 0);
            tbf.pinvCb.setSelectedIndex(0);
        }
    }

    /**
     * Adds the MCTS-Sets to the combobox
     */
    public void addMctsToComboBox() {
        if(mctsResults == null)
            mctsResults = getMctsResults();
        if(!mctsResults.isEmpty()) {
            tbf.mctsCb.removeAllItems();
            tbf.mctsCb.setEnabled(true);

            String[] configName;
            String withOutOrWith;
            for(Map.Entry<Configuration, Result> entry : mctsResults.entrySet()) {
                Configuration config = entry.getKey();
                Mcts results = project.getResult(new MctsTool(), config);                
                if(results != null) {
                    configName = config.toString(strings).split(" ");
                    if(configName[2].contains("with"))
                        withOutOrWith = configName[2];
                    else
                        withOutOrWith = configName[3];
                    tbf.mctsCb.addItem(configName[0].split("-")[0]+" ("+withOutOrWith+" trivial T-inv)");
                    Iterator it = results.iterator();
                    MctsWrapper wrapper;
                    while(it.hasNext()) {
                        wrapper = new MctsWrapper((TInvariant)it.next());
                        if(!wrapper.getMcts().isEmpty()) {
                            tbf.mctsCb.addItem(wrapper);
                        }
                    }
                }
            }            
            tbf.mctsCb.setSelectedIndex(0);
        }
    }

    /**
     * Reset the coloring of all vertex to default
     */
    private void resetColor() {
        for(NetViewerNode n : g.getVertices()) {
            if(n.getNodeType().equalsIgnoreCase(NetViewer.TRANSITION))
                n.setColor(Color.BLACK);
            if(n.getNodeType().equalsIgnoreCase(NetViewer.PLACE)) {
                if(n.isLogical())
                    n.setColor(Color.LIGHT_GRAY);
                else if(n.isMasterNode())
                    n.setColor(Color.DARK_GRAY);
                else
                    n.setColor(Color.WHITE);
            }
        }
        for(NetViewerEdge e : g.getEdges()) {
            if(e.getVisible())
                e.setColorForAllEdges(Color.BLACK);
        }

        vv.repaint();
    }

    /**
     * Add or delete (or update) a node to the search bar
     */
    protected void updateSearchBar(Collection <NetViewerNode> vertices) {

        List<NetViewerNode> places = new ArrayList<NetViewerNode>();
        List<NetViewerNode> transitions = new ArrayList<NetViewerNode>();

        List<String> allPlaceNames = new ArrayList<String>();
        List<String> doublePlaceNames = new ArrayList<String>();
        List<String> allTransitionNames = new ArrayList<String>();
        List<String> doubleTransitionNames = new ArrayList<String>();

        for(NetViewerNode nvNode : vertices) {
            if(!nvNode.isLogical() && !nvNode.getNodeType().equals(BEND)) {
                if(nvNode.getNodeType().equals(PLACE)) {
                    places.add(nvNode);
                    if(!allPlaceNames.contains(nvNode.getName()))
                        allPlaceNames.add(nvNode.getName());
                    else
                        doublePlaceNames.add(nvNode.getName());
                }
                else {
                    transitions.add(nvNode);
                    if(!allTransitionNames.contains(nvNode.getName()))
                        allTransitionNames.add(nvNode.getName());
                    else
                        doubleTransitionNames.add(nvNode.getName());
                }
            }
        }

        Collections.sort(places, new MyStringComparator());
        Collections.sort(transitions, new MyStringComparator());

        allPlacesModel.clear();
        allTransitionsModel.clear();

        Boolean doublePlaces = false, doubleTransitions = false;

        for(NetViewerNode nvNode : places) {
            allPlacesModel.addElement(nvNode);
            if(doublePlaceNames.contains(nvNode.getName())) {
                nvNode.setSerachBarColor(Color.RED);
                doublePlaces = true;
            }
            else
                nvNode.setSerachBarColor(Color.BLACK);
        }

        if(doublePlaces)
            sb.doublePlacesWarning.setText(strings.get("NVWarningDoublePlaceNames"));
        else
            sb.doublePlacesWarning.setText("");

        for(NetViewerNode nvNode : transitions) {
            allTransitionsModel.addElement(nvNode);
            if(doubleTransitionNames.contains(nvNode.getName())) {
                nvNode.setSerachBarColor(Color.RED);
                doubleTransitions = true;
            }
            else
                nvNode.setSerachBarColor(Color.BLACK);
        }
        
        if(doubleTransitions)
            sb.doubleTransitionsWarning.setText(strings.get("NVWarningDoubleTransitionNames"));
        else
            sb.doubleTransitionsWarning.setText("");

        sb.placesLabel.setText(strings.get("NVPlaces")+" ("+allPlacesModel.getSize()+")");
        sb.transitionsLabel.setText(strings.get("NVTransitions")+" ("+allTransitionsModel.getSize()+")");
    }

    private void updateInfoBar() {
        int placeCounter = project.getPetriNet().places().size();
        int transitionCounter = project.getPetriNet().transitions().size();
        int edgeCounter = 0;

        for(Transition transition : project.getPetriNet().transitions()) {
            edgeCounter+=transition.inputs().size();
            edgeCounter+=transition.outputs().size();
        }

        infoBarLabel.setText(" Places: "+placeCounter+" | Transitions: "+transitionCounter+" | Edges: "+edgeCounter);
    }


    /**
     * Are T-Invariants loaded?
     * @return
     */
    public boolean hasTinvs() {
        if(tbf.allInvCb.getItemCount() == 0)
            return false;
        else
            return true;
    }

    /**
     * Are MCT-Sets loaded?
     * @return
     */
    public boolean hasMcts() {
        if(tbf.mctsCb.getItemCount() == 0)
            return false;
        else
            return true;
    }

    protected void saveFromShortKey() throws IOException, ClassNotFoundException {
        saveView(false, true);
    }

    /**
     * Saves the current view
     */
    public void saveView(Boolean shortWay, Boolean saveAs) throws IOException, ClassNotFoundException {
        // Write out the layout
        File tmpFileLayout = File.createTempFile("MLtmp", ".tmp", FileUtils.getTempDir());
        layout.persist(tmpFileLayout.getAbsolutePath());
        // Read the layout and write it to a byte array
        BufferedInputStream bis = new BufferedInputStream(new FileInputStream(tmpFileLayout));
        long size = tmpFileLayout.length();
        byte[] layoutData = new byte[(int)size];
        int by, i=0;
        while((by=bis.read()) != -1)
            layoutData[i++] = (byte)by;

        // Write the graph to a byte array
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        ObjectOutputStream oos = new ObjectOutputStream(bos);
        oos.writeObject(g);
        oos.flush();
        oos.close();
        bos.close();
        byte [] graphData = bos.toByteArray();

        // Write the results to a byte array
        bos = new ByteArrayOutputStream();
        oos = new ObjectOutputStream(bos);
        oos.writeObject(project.getAllResults());
        oos.flush();
        oos.close();
        bos.close();
        byte [] resultsData = bos.toByteArray();

        // Write the results to a byte array
        View view = null;
        
        if(saveAs && project.getViewWrapper().getViewCounter() > 0) {
            view = new View(((View)tbf.viewsCb.getSelectedItem()).toString(), layoutData, graphData, resultsData);
            project.getViewWrapper().replaceView((View)tbf.viewsCb.getSelectedItem(), view);
            int index = tbf.viewsCb.getSelectedIndex();
            tbf.viewsCb.insertItemAt(view, index);
            tbf.viewsCb.removeItemAt(index+1);
            tbf.viewsCb.setSelectedItem(view);
        }
        else {
            view = new View("View "+(viewCounter++), layoutData, graphData, resultsData);
            project.getViewWrapper().addView(view);
            updateViewsCb();
            tbf.viewsCb.addItem(view);
            tbf.viewsCb.setSelectedItem(view);
            if(!shortWay)
                showEditView();
        }
        
        if(project.save()) {
            try {
                mainDialog.projectLoaded();
            } catch (InterruptedException ex) {
                Logger.getLogger(NetViewer.class.getName()).log(Level.SEVERE, null, ex);
            }
        }

        tmpFileLayout.delete();
        udm.clearUndoHistroy();
    }

    protected void exportView() throws IOException {
        JFileChooser viewFileChooser = new JFileChooser();
        viewFileChooser.setFileFilter(new MyFileFilter("view", "MonaLisa view file"));

        viewFileChooser.setDialogTitle(strings.get("ViewFileLocation"));
        if (viewFileChooser.showDialog(this, strings.get("NVSave")) != JFileChooser.APPROVE_OPTION)
            return;

        File viewFile = viewFileChooser.getSelectedFile();
        viewFile = ((MyFileFilter)viewFileChooser.getFileFilter()).checkFileNameForExtension(viewFile);
        ObjectOutputStream output = new ObjectOutputStream(new FileOutputStream(viewFile));
        output.writeObject((View)tbf.viewsCb.getSelectedItem());
        output.close();
    }

    protected void importView() throws FileNotFoundException, IOException, ClassNotFoundException {
        JFileChooser viewFileChooser = new JFileChooser();
        viewFileChooser.setFileFilter(new MyFileFilter("view", "MonaLisa view file"));

        viewFileChooser.setDialogTitle(strings.get("ViewFileLocation"));
        if (viewFileChooser.showOpenDialog(this) != JFileChooser.APPROVE_OPTION)
            return;

        File viewFile = viewFileChooser.getSelectedFile();        
        ObjectInputStream input = new ObjectInputStream(new FileInputStream(viewFile));
        View view = (View)input.readObject();        
        input.close();
        project.getViewWrapper().addView(view);
        tbf.viewsCb.addItem(view);
        tbf.viewsCb.setSelectedItem(view);

        updateViewsCb();        
        netChanged();
    }

    /**
     * Delete a view from the list
     * @throws IOException
     */
    protected void deleteView() throws IOException {
        if(tbf.viewsCb.getSelectedItem().getClass().getSimpleName().equalsIgnoreCase("View")) {
            View view = (View)tbf.viewsCb.getSelectedItem();
            project.getViewWrapper().removeView(view);
            tbf.viewsCb.removeItem(view);
            updateViewsCb();
            project.save();
        }
        project.setProjectChanged(true);
    }

    /**
     * Updates the views ComboBox and disable then, if all views deleted
     */
    private void updateViewsCb() {
        int nbrOfViews = project.getViewWrapper().getViewCounter();
        if(nbrOfViews == 0) {
            tbf.saveViewButton.setEnabled(false);
            tbf.viewsCb.setEnabled(false);
            tbf.viewsCb.removeAllItems();
            tbf.viewsCb.addItem(strings.get("NVNoViews"));
            tbf.deleteViewButton.setEnabled(false);
            tbf.editViewButton.setEnabled(false);
            tbf.exportViewButton.setEnabled(false);
        }
        else {
            tbf.viewsCb.setEnabled(true);
            tbf.deleteViewButton.setEnabled(true);
            tbf.editViewButton.setEnabled(true);
            tbf.saveViewButton.setEnabled(true);
            tbf.importViewButton.setEnabled(true);
            tbf.exportViewButton.setEnabled(true);
        }
    }

    public View getSelectedView() {
        return (View)tbf.viewsCb.getSelectedItem();
    }

    /**
     * Save the properties of selected vertex
     */
    protected void saveVertexSetup() {
        if(vv.getPickedVertexState().getPicked().size() == 1) {
            NetViewerNode nvNode = (NetViewerNode) vv.getPickedVertexState().getPicked().toArray()[0];
            udm.addEvent(new UndoEvent(UndoEvent.SAVE_VERTEX, nvNode));
            saveVertexSetup(nvNode);
        } else { 
            // If a logical node is selected and edited from the searchbar, this part is needed
            NetViewerNode nvNode = ((NetViewerNode)vv.getPickedVertexState().getPicked().toArray()[0]).getMasterNode();
            udm.addEvent(new UndoEvent(UndoEvent.SAVE_VERTEX, nvNode));
            saveVertexSetup(nvNode);            
        }        
        project.setProjectChanged(true);
    }
   
    /**
     * Save the properties for a given vertex
     * @param nvNode
     */
    private void saveVertexSetup(NetViewerNode nvNode) {
        int corners = 0;
        if(vsf.circleRadioButon.isSelected())
            corners = 0;
        else if(vsf.polygoneRadioButton.isSelected())
            corners = (Integer)vsf.cornerSpinner.getValue();

        String oldName = nvNode.getName();
        String newName = vsf.vertexNameTextField.getText();
        nvNode.getMasterNode().setColorForAllNodes(vsf.showColorLabelVertex.getForeground());
        nvNode.getMasterNode().setCornersForAllNodes(corners);
        nvNode.getMasterNode().setNameForAllNodes(newName);

        name2Node.remove(oldName);
        name2Node.put(newName, nvNode);

        if(nvNode.getNodeType().equalsIgnoreCase(NetViewer.PLACE)) {
            Place place = project.getPetriNet().findPlace(nvNode.getMasterNode().getId());
            place.putProperty("name", nvNode.getName());
        }
        else if(nvNode.getNodeType().equalsIgnoreCase(NetViewer.TRANSITION)) {
            Transition transition = project.getPetriNet().findTransition(nvNode.getId());
            transition.putProperty("name", nvNode.getName());
        }

        if(vsf.vertexNoteTextArea.getText() != "")
            nvNode.putProperty("toolTip", vsf.vertexNoteTextArea.getText());

        if(vsf.C.isSelected())
            nvNode.setLabelPosition(Position.CNTR);
        else if(vsf.E.isSelected())
            nvNode.setLabelPosition(Position.E);
        else if(vsf.N.isSelected())
            nvNode.setLabelPosition(Position.N);
        else if(vsf.NE.isSelected())
            nvNode.setLabelPosition(Position.NE);
        else if(vsf.NW.isSelected())
            nvNode.setLabelPosition(Position.NW);
        else if(vsf.S.isSelected())
            nvNode.setLabelPosition(Position.S);
        else if(vsf.SE.isSelected())
            nvNode.setLabelPosition(Position.SE);
        else if(vsf.SW.isSelected())
            nvNode.setLabelPosition(Position.SW);
        else if(vsf.W.isSelected())
            nvNode.setLabelPosition(Position.W);

        vv.repaint();
        updateSearchBar(g.getVertices());
        project.setProjectChanged(true);

        vsf.setVisible(false);
        this.setEnabled(true);
    }

    /**
     * Save the properties for a  set of vertices
     */
    protected void saveVerticesSetup() {
        for(NetViewerNode nvNode : selection4Modification) {
            if(!nvNode.getNodeType().equalsIgnoreCase(NetViewer.BEND)) {
                nvNode.getMasterNode().setColorForAllNodes(vsf.showColorLabelVertex.getForeground());
                
                if(vsf.C.isSelected())
                    nvNode.setLabelPosition(Position.CNTR);
                else if(vsf.E.isSelected())
                    nvNode.setLabelPosition(Position.E);
                else if(vsf.N.isSelected())
                    nvNode.setLabelPosition(Position.N);
                else if(vsf.NE.isSelected())
                    nvNode.setLabelPosition(Position.NE);
                else if(vsf.NW.isSelected())
                    nvNode.setLabelPosition(Position.NW);
                else if(vsf.S.isSelected())
                    nvNode.setLabelPosition(Position.S);
                else if(vsf.SE.isSelected())
                    nvNode.setLabelPosition(Position.SE);
                else if(vsf.SW.isSelected())
                    nvNode.setLabelPosition(Position.SW);
                else if(vsf.W.isSelected())
                    nvNode.setLabelPosition(Position.W);                   
            }
        }
        selection4Modification.clear();

        vsf.vertexTypeLabel.setEnabled(true);
        vsf.vertexNameTextField.setEnabled(true);
        vsf.vertexNoteTextArea.setEnabled(true);

        vv.repaint();
        project.setProjectChanged(true);
        
        vsf.setVisible(false);
        this.setEnabled(true);
    }

   /**
     * Save the properties of the selected edge
     */
    protected void saveEdgeSetup() {
        if(vv.getPickedEdgeState().getPicked().size() == 1) {
            NetViewerEdge nvEdge = ((NetViewerEdge) vv.getPickedEdgeState().getPicked().toArray()[0]).getMasterEdge();
            udm.addEvent(new UndoEvent(UndoEvent.SAVE_EDGE, nvEdge, nvEdge.getWeight()));
            saveEdgeSetup(nvEdge);
        } else {
            saveEdgesSetup(vv.getPickedEdgeState().getPicked());
        }
    }

    /**
     * Save the properties of a given egde
     * @param nvEdge
     */
    private void saveEdgeSetup(NetViewerEdge nvEdge) {
        if(nvEdge.getWeight() != (Integer)esf.weightSpinner.getValue()) {
            netChanged = true;
            netChanged();
        }

        nvEdge.setWeightForAllEdges((Integer)esf.weightSpinner.getValue());
        nvEdge.setColorForAllEdges(esf.showColorLabelEdge.getForeground());

        if(nvEdge.getSource().getNodeType().equalsIgnoreCase(NetViewer.PLACE)) {
            Place from = project.getPetriNet().findPlace(nvEdge.getSource().getMasterNode().getId());
            Transition to = project.getPetriNet().findTransition(nvEdge.getAim().getId());
            project.getPetriNet().getArc(from, to).putProperty("weight", nvEdge.getWeight());
        }
        else if(nvEdge.getSource().getNodeType().equalsIgnoreCase(NetViewer.TRANSITION)) {
            Transition from = project.getPetriNet().findTransition(nvEdge.getSource().getId());
            Place to = project.getPetriNet().findPlace(nvEdge.getAim().getMasterNode().getId());
            project.getPetriNet().getArc(from, to).putProperty("weight", nvEdge.getWeight());
        }

        if(vsf.vertexNoteTextArea.getText() != "")
            nvEdge.putProperty("toolTip", esf.edgeNoteTextArea.getText());

        vv.repaint();
        project.setProjectChanged(true);
        updateSearchBar(g.getVertices());
        esf.setVisible(false);
        this.setEnabled(true);
    }

    /**
     * Save the properties of a set of given egdes
     * @param nvEdge
     */
    private void saveEdgesSetup(Set<NetViewerEdge> nvEdges) {
        List<NetViewerEdge> worked = new ArrayList<NetViewerEdge>();
        NetViewerEdge nvEdge;        
        for(NetViewerEdge e : nvEdges) {
            nvEdge = e.getMasterEdge();
            if(!worked.contains(nvEdge)) {
                worked.add(nvEdge);
                
                if(nvEdge.getWeight() != (Integer)esf.weightSpinner.getValue()) {
                    netChanged = true;
                    netChanged();
                }

                nvEdge.setWeightForAllEdges((Integer)esf.weightSpinner.getValue());
                nvEdge.setColorForAllEdges(esf.showColorLabelEdge.getForeground());

                if(nvEdge.getSource().getNodeType().equalsIgnoreCase(NetViewer.PLACE)) {
                    Place from = project.getPetriNet().findPlace(nvEdge.getSource().getMasterNode().getId());
                    Transition to = project.getPetriNet().findTransition(nvEdge.getAim().getId());
                    project.getPetriNet().getArc(from, to).putProperty("weight", nvEdge.getWeight());
                }
                else if(nvEdge.getSource().getNodeType().equalsIgnoreCase(NetViewer.TRANSITION)) {
                    Transition from = project.getPetriNet().findTransition(nvEdge.getSource().getId());
                    Place to = project.getPetriNet().findPlace(nvEdge.getAim().getMasterNode().getId());
                    project.getPetriNet().getArc(from, to).putProperty("weight", nvEdge.getWeight());
                }

                if(vsf.vertexNoteTextArea.getText() != "")
                    nvEdge.putProperty("toolTip", esf.edgeNoteTextArea.getText());
            }
        }
        
        vv.repaint();
        updateSearchBar(g.getVertices());
        project.setProjectChanged(true);
        
        esf.setVisible(false);
        this.setEnabled(true);
    }        
    
    /**
     * Create a new logical place of selected place
     */
    protected void createLogicalPlace() {
        NetViewerNode sourceNode = (NetViewerNode) vv.getPickedVertexState().getPicked().toArray()[0];
        Object[] selectedNodes = lpf.neighborsList.getSelectedValues();
        createLogicalPlace(sourceNode, selectedNodes, false, null);
    }

    /**
     * Create a new logical place
     * @param sourceNode
     * @param selectedNodes
     * @param undo
     * @param nodePosition
     */
    private void createLogicalPlace(NetViewerNode sourceNode, Object[] selectedNodes, Boolean undo, Point2D nodePosition) {
        if(selectedNodes.length > 0) {
            NetViewerNode newNode = sourceNode.getMasterNode().generateLocicalPlace(getNewNodeId());
            sourceNode.getMasterNode().setColor(Color.GRAY);
            g.addVertex(newNode);

            if(!(sourceNode.getMasterNode().getColor().equals(Color.WHITE) || sourceNode.getMasterNode().getColor().equals(Color.GRAY)))
                newNode.setColor(sourceNode.getColor());
            
            if(sourceNode.getCorners() != 0)
                newNode.setCorners(sourceNode.getCorners());

            NetViewerEdge oldEdge, newEdge;
            int i = 1;
            for(Object n : selectedNodes) {
                if(n == null) {
                    continue;
                }                
                // in which direction are the edge?
                oldEdge = g.findEdge(sourceNode, (NetViewerNode)n);
                if(oldEdge != null) {
                    newEdge = new NetViewerEdge("L"+newEdgesCounter+(int)(Math.random() * (1000 - 1) + 1), oldEdge.getWeight(), newNode, (NetViewerNode)n);
                    g.removeEdge(oldEdge);
                    g.addEdge(newEdge, newEdge.getSource(), newEdge.getAim(), EdgeType.DIRECTED);

                    sourceNode.removeOutEdge(oldEdge);
                }
                else {
                    oldEdge = g.findEdge((NetViewerNode)n, sourceNode);
                    newEdge = new NetViewerEdge("L"+newEdgesCounter+(int)(Math.random() * (1000 - 1) + 1), oldEdge.getWeight(), (NetViewerNode)n, newNode);
                    removeBendEdges(oldEdge);
                    g.removeEdge(oldEdge);
                    g.addEdge(newEdge, newEdge.getSource(), newEdge.getAim(), EdgeType.DIRECTED);
                    sourceNode.removeInEdge(oldEdge);
                }
                newEdgesCounter++;

                oldEdge.getAim().removeInEdge(oldEdge);
                oldEdge.getSource().removeOutEdge(oldEdge);

                i++;
                if(i == lpf.neighborsList.getModel().getSize())
                    break;
            }
            if(nodePosition == null)
                layout.setLocation(newNode, calcNewPoint(layout.transform(sourceNode)));
            else
                layout.setLocation(newNode, nodePosition);
            vv.repaint();

            if(!undo)
                udm.addEvent(new UndoEvent(UndoEvent.ADD_LOGICAL_PLACE, newNode, layout.transform(newNode)));
        }

        lpf.setVisible(false);
        vv.getPickedEdgeState().clear();
        neighborsListModel.removeAllElements();
        this.setEnabled(true);
        project.setProjectChanged(true);
    }

    /**
     * Create all logical places of the selected place
     */
    protected void generateAllLocicalPlaces() {
        NetViewerNode sourceNode = (NetViewerNode) vv.getPickedVertexState().getPicked().toArray()[0];
        Object[] selectedNodes = new Object[1];
        for(int i = 0; i < lpf.neighborsList.getModel().getSize()-1; i++) {
            selectedNodes[0] = lpf.neighborsList.getModel().getElementAt(i);
            createLogicalPlace(sourceNode, selectedNodes, false, null);
        }
    }

    /**
     * Remove all bends of a gives edge
     * @param bendEdges
     */
    private void removeBendEdges(NetViewerEdge oldEdge) {
        List<NetViewerEdge> edgeList = oldEdge.getMasterEdge().getBendEdges();

        for(NetViewerEdge edge : edgeList) {
            edge.getSource().removeOutEdge(edge);
            edge.getAim().removeInEdge(edge);

            g.removeEdge(edge);

            if(edge.getAim().getNodeType().equalsIgnoreCase(NetViewer.BEND) )
                g.removeVertex(edge.getAim());
        }
        oldEdge.getMasterEdge().clearBendEdges();
        oldEdge.getMasterEdge().setVisible(true);

        oldEdge = null;
        project.setProjectChanged(true);
    }

    /**
     * Remove a logical place
     */
    protected void removeLogicalPlace() {
        removeLogicalPlace((NetViewerNode) vv.getPickedVertexState().getPicked().toArray()[0], null, false);
    }

    protected void removeLogicalPlaces(List<NetViewerNode> toDelete) {
        if(toDelete.size() == 2) {
            if(toDelete.get(0).isMasterNode())
                removeLogicalPlace(toDelete.get(1), toDelete.get(0), false);
            else
                removeLogicalPlace(toDelete.get(0), toDelete.get(1), false);
        }
        else {
            NetViewerNode toMelt = toDelete.get(0);
            for(NetViewerNode nvNode : toDelete.subList(1, toDelete.size())) 
                removeLogicalPlace(nvNode, toMelt, false);            
        }
        vv.getRenderContext().getPickedVertexState().clear();
        project.setProjectChanged(true);
    }

    /**
     * Removes a logical places. All egdes a go back to master node or to a other given logical places
     * @param nvNode Place to remove
     * @param aim Aim for edges. null : master node
     * @param undo is the function call from the undo manager?
     */
    private void removeLogicalPlace(NetViewerNode nvNode, NetViewerNode aim, Boolean undo) {
        if(nvNode.equals(nvNode.getMasterNode()))
            return;
        NetViewerEdge newEdge, edgeToDelete;
        List<NetViewerNode> selectedNodes = new ArrayList<NetViewerNode>();
        List<NetViewerEdge> deletedEdges = new ArrayList<NetViewerEdge>();

        if(aim == null)
            aim = nvNode.getMasterNode();

        for(NetViewerEdge e : nvNode.getInEdges()) {
            newEdge = new NetViewerEdge("N"+newEdgesCounter+(int)(Math.random() * (1000 - 1) + 1), e.getWeight(), e.getSource(), aim);
            g.addEdge(newEdge, newEdge.getSource(), newEdge.getAim(), EdgeType.DIRECTED);
            edgeToDelete = g.findEdge(e.getSource(), nvNode);
            deletedEdges.add(edgeToDelete);
            g.removeEdge(edgeToDelete);
            selectedNodes.add(newEdge.getSource());
            newEdgesCounter++;
        }
        for(NetViewerEdge e : nvNode.getOutEdges()) {
            newEdge = new NetViewerEdge("N"+newEdgesCounter+(int)(Math.random() * (1000 - 1) + 1), e.getWeight(), aim, e.getAim());
            g.addEdge(newEdge, newEdge.getSource(), newEdge.getAim(), EdgeType.DIRECTED);
            edgeToDelete = g.findEdge(nvNode, e.getAim());
            g.removeEdge(edgeToDelete);
            deletedEdges.add(edgeToDelete);
            selectedNodes.add(newEdge.getAim());
            newEdgesCounter++;
        }
        
        for(NetViewerEdge nvEdge : deletedEdges) {
            nvEdge.getSource().removeOutEdge(nvEdge);
            nvEdge.getAim().removeInEdge(nvEdge);
        }

        g.removeVertex(nvNode);
        nvNode.getMasterNode().removeLocicalNode(nvNode);

        if(!nvNode.getMasterNode().getColor().equals(Color.WHITE))
             nvNode.getMasterNode().setColor(nvNode.getMasterNode().getColor());
        else
            nvNode.getMasterNode().setColor(Color.WHITE);

        if(nvNode.getMasterNode().getLogicalPlaces().size() == 1)
            nvNode.getMasterNode().setColor(Color.WHITE);

        vv.repaint();
        project.setProjectChanged(true);

        if(!undo)
            udm.addEvent(new UndoEvent(UndoEvent.REMOVE_LOGICAL_PLACE, aim, selectedNodes.toArray()));
    }

    /**
     * Insert a bend into selected edge
     */
    protected void insertBend(int x, int y) {
        NetViewerEdge oldEdge = (NetViewerEdge) vv.getPickedEdgeState().getPicked().toArray()[0];
        insertBend(oldEdge, x, y, false);
    }

    /**
     * Insert a bend in a given edge
     * @param oldEdge
     * @param undo
     */
    protected void insertBend(NetViewerEdge oldEdge, int x, int y, Boolean undo) {
        NetViewerEdge newEdge1, newEdge2;
        NetViewerNode newNode = new NetViewerNode(getNewNodeId(), NetViewer.BEND, "B"+bendCounter);
        NetViewerEdge masterEdge = oldEdge.getMasterEdge();
        
        g.addVertex(newNode);

        newEdge1 = new NetViewerEdge("B"+(++bendCounter)+(int)(Math.random() * (1000 - 1) + 1), oldEdge.getWeight(), oldEdge.getSource(), newNode, masterEdge);
        g.addEdge(newEdge1, newEdge1.getSource(), newEdge1.getAim(), EdgeType.UNDIRECTED);

        newEdge2 = new NetViewerEdge("B"+(++bendCounter)+(int)(Math.random() * (1000 - 1) + 1), oldEdge.getWeight(), newNode, oldEdge.getAim(), masterEdge);
        if(newEdge2.getAim().getNodeType().equalsIgnoreCase(NetViewer.BEND) )
            g.addEdge(newEdge2, newEdge2.getSource(), newEdge2.getAim(), EdgeType.UNDIRECTED);
        else
            g.addEdge(newEdge2, newEdge2.getSource(), newEdge2.getAim(), EdgeType.DIRECTED);

        if(!oldEdge.getSource().getNodeType().equalsIgnoreCase(NetViewer.BEND) && !oldEdge.getAim().getNodeType().equalsIgnoreCase(NetViewer.BEND))
            oldEdge.setVisible(false);
        else {
            g.removeEdge(oldEdge);
            oldEdge.getMasterEdge().removeBendEdge(oldEdge);
            oldEdge.getAim().removeInEdge(oldEdge);
            oldEdge.getSource().removeOutEdge(oldEdge);
        }

        layout.setLocation(newNode, vv.getRenderContext().getMultiLayerTransformer().inverseTransform(calcNewPoint(new Point(x,y))));
        vv.repaint();

        if(!undo)
            udm.addEvent(new UndoEvent(UndoEvent.INSERT_BEND, newEdge1));
        project.setProjectChanged(true);
    }

    /**
     * Remove a bend from a edge
     */
    protected void removeBend() {
        removeBend((NetViewerEdge) vv.getPickedEdgeState().getPicked().toArray()[0], false);
    }

    /**
     * Remove a bend from a given edge
     * @param edge
     */
    protected void removeBend(NetViewerEdge edge, Boolean undo) {
        NetViewerEdge invisibleEdge = null, otherEdge = null, newEdge = null;

        // NetViewer.BEND ------ NODE
        if(edge.getSource().getNodeType().equalsIgnoreCase(NetViewer.BEND) && !edge.getAim().getNodeType().equalsIgnoreCase(NetViewer.BEND)) {
            edge.getAim().removeInEdge(edge);
            edge.getSource().removeOutEdge(edge);
            otherEdge = (NetViewerEdge)edge.getSource().getInEdges().toArray()[0];
            // NODE ------ NetViewer.BEND
            if(!otherEdge.getSource().getNodeType().equalsIgnoreCase(NetViewer.BEND)) 
                invisibleEdge = g.findEdge(otherEdge.getSource(), edge.getAim());
            // NODE ------ NODE
            else {
                newEdge = new NetViewerEdge("B"+(++bendCounter)+(int)(Math.random() * (1000 - 1) + 1), 0, otherEdge.getSource(), edge.getAim(), edge.getMasterEdge());
                g.addEdge(newEdge, newEdge.getSource(), newEdge.getAim(), EdgeType.DIRECTED);
            }
            g.removeVertex(edge.getSource());
        }
        // NODE ------ NetViewer.BEND
        else if(edge.getAim().getNodeType().equalsIgnoreCase(NetViewer.BEND) && !edge.getSource().getNodeType().equalsIgnoreCase(NetViewer.BEND)) {
            edge.getSource().removeOutEdge(edge);
            edge.getAim().removeInEdge(edge);
            otherEdge = (NetViewerEdge)edge.getAim().getOutEdges().toArray()[0];

            // NetViewer.BEND ------ NODE
            if(!otherEdge.getAim().getNodeType().equalsIgnoreCase(NetViewer.BEND))
                invisibleEdge = g.findEdge(edge.getSource(), otherEdge.getAim());
            // NetViewer.BEND ------ NetViewer.BEND
            else {
                newEdge = new NetViewerEdge("B"+(++bendCounter)+(int)(Math.random() * (1000 - 1) + 1), 0, edge.getSource(), otherEdge.getAim(), edge.getMasterEdge());
                g.addEdge(newEdge, newEdge.getSource(), newEdge.getAim(), EdgeType.UNDIRECTED);
            }
            g.removeVertex(edge.getAim());
        }
        // NetViewer.BEND ------ NetViewer.BEND
        else {
            edge.getAim().removeInEdge(edge);
            edge.getSource().removeOutEdge(edge);
            otherEdge = ((NetViewerEdge) edge.getSource().getInEdges().toArray()[0]);
            otherEdge.getSource().removeOutEdge(otherEdge);
            
            newEdge = new NetViewerEdge("B"+(++bendCounter)+(int)(Math.random() * (1000 - 1) + 1), 0, otherEdge.getSource(), edge.getAim(), edge.getMasterEdge());
            g.removeVertex(otherEdge.getAim());            
            g.addEdge(newEdge, newEdge.getSource(), newEdge.getAim(), EdgeType.UNDIRECTED);
        }

        otherEdge.getMasterEdge().removeBendEdge(otherEdge);
        g.removeEdge(otherEdge);
        edge.getMasterEdge().removeBendEdge(edge);
        g.removeEdge(edge);

        if(invisibleEdge != null) 
            invisibleEdge.setVisible(true);    

        edge = null;
        vv.repaint();
        project.setProjectChanged(true);
    }

    /**
     * Melting two or more places
     */
    protected void meltPlaces() {
        melt(sb.allPlacesList.getSelectedValues());
    }

    /**
     * Melting two or more places
     */
    protected void meltTransitions() {
        melt(sb.allTransitionsList.getSelectedValues());
    }

    /**
     * Melts a list of given places/transitions
     * @param selected
     */
    private void melt(Object[] selected) {
        // Define the place which "survive"
        NetViewerNode meltingPot = (NetViewerNode)selected[0];

        udm.addEvent(new UndoEvent(UndoEvent.MELTING_VETICES, selected));

        // Go through all other vertices to melt these
        int lenSelected = selected.length;
        NetViewerNode toMelt = null;
        NetViewerEdge newEdge = null;
        for(int i = 1; i < lenSelected; i++) {
            toMelt = (NetViewerNode)selected[i];

            for(NetViewerEdge e : toMelt.getInEdges()) {
                newEdge = new NetViewerEdge("e"+(++edgeCount), e.getWeight(), e.getSource(), meltingPot);
                e.getSource().removeOutEdge(e);
                g.removeEdge(e);
                removeArcFromPetriNet(e);
                g.addEdge(newEdge, newEdge.getSource(), newEdge.getAim(), EdgeType.DIRECTED);
                addArcToPetriNet(newEdge);
            }

            for(NetViewerEdge e : toMelt.getOutEdges()) {
                newEdge = new NetViewerEdge("e"+(++edgeCount), e.getWeight(), meltingPot, e.getAim());
                e.getAim().removeInEdge(e);
                g.removeEdge(e);
                removeArcFromPetriNet(e);
                g.addEdge(newEdge, newEdge.getSource(), newEdge.getAim(), EdgeType.DIRECTED);
                addArcToPetriNet(newEdge);
            }

            if(toMelt.getNodeType().equals(NetViewer.TRANSITION))
                transitionMap.remove(toMelt.getId());
            else
                placeMap.remove(toMelt.getId());

            g.removeVertex(toMelt);
            removeVertexFromPetriNet(toMelt);
        }

        updateSearchBar(g.getVertices());
        updateInfoBar();
        vv.repaint();
        project.setProjectChanged(true);
    }

    /**
     * Calculates a new point, 150 pixel around the given point
     * @param oldPoint
     * @return
     */
    private Point2D calcNewPoint(Point2D oldPoint) {
        Point.Double newPoint = new Point.Double();
        newPoint.x = oldPoint.getX() + 30;
        newPoint.y = oldPoint.getY() + 30;
        return newPoint;
    }

    /**
     * Colors all transitions which occurrs in one or more T-Invariants
     */
    private void colorCTITransitions() {
        if(tinvs != null) {
            resetColor();

            Iterator<Transition> it;
            for(TInvariant tinv: tinvs) {
                it = tinv.transitions().iterator();
                while(it.hasNext()) {
                        transitionMap.get(it.next().id()).setColor(Color.YELLOW);
                }
            }
        }
        vv.repaint();
    }

    /**
     * Counts occurence of transitions in T-Invariants and generate a heatmap
     */
    private void colorImportantTransitions(Boolean byOccurrence) {
        if(tinvs != null) {
            float min = 0, max = 0, maxmin = 0;
            float[] hsbvals = Color.RGBtoHSB(heatMapColor.getRed(), heatMapColor.getGreen(), heatMapColor.getBlue(), null);
            float norm = 0;
            Map<Transition, Integer> allFactors = new HashMap<Transition, Integer>();

            resetColor();

            Iterator<Transition> it;
            Transition t;
            for(TInvariant tinv : tinvs) {
                it = tinv.transitions().iterator();
                while(it.hasNext()) {
                    t = it.next();
                    if(!allFactors.containsKey(t))
                        allFactors.put(t, 0);
                    if(byOccurrence)
                        allFactors.put(t, allFactors.get(t) + 1);
                    else
                        allFactors.put(t, allFactors.get(t) + tinv.factor(t));
                }
            }

            Iterator<Integer> itInt = allFactors.values().iterator();
            int value;
            while(itInt.hasNext()) {
                value = itInt.next();
                if(value < min)
                    min = value;
                if(value > max)
                    max = value;
            }
            maxmin = max-min;

            messageLabel.setText(strings.get("NVImportantTransitionMessage", (int)min, (int)max));

            Iterator<Entry<Transition, Integer>> itEntry = allFactors.entrySet().iterator();
            Entry<Transition, Integer> entry;
            while(itEntry.hasNext()) {
                entry = itEntry.next();
                norm = (entry.getValue() - min) / maxmin;
                if(norm < 0.05)
                    norm = (float)0.05;
                transitionMap.get(entry.getKey().id()).setColor(new Color(Color.HSBtoRGB(hsbvals[0], norm, hsbvals[2])));
            }
            vv.repaint();
        }
    }

    /**
     * Hide label from a given node
     * @param nvNode
     */
    protected void hideLabel(NetViewerNode nvNode) {
        nvNode.setShowLabel(!nvNode.showLabel());
        vv.repaint();
    }

    /**
     * Knock out a transition and show the new Petri net
     * @param nvNode
     * @param mode  0=Standart, 1=heatmap by occurrence, 2=heatmap by factor
     */
    protected void knockOut(List<NetViewerNode> nvNodes, int mode) {
        List<TInvariant> notKnockedOutTinvariants = new ArrayList<TInvariant>();
        List<Transition> notKnockedOutTransitions = new ArrayList<Transition>();
        Map<Transition, Integer> factors = new HashMap<Transition, Integer>();
        int factor;
        float min = 0, max = 0, maxmin = 0, norm = 0;
        double percentTransitions = 100.0, percentInvariants = 0.0;
        float[] hsbvals = Color.RGBtoHSB(notKnockedOutColor.getRed(), notKnockedOutColor.getGreen(), notKnockedOutColor.getBlue(), null);
        Boolean isKnockedOut;

        for(TInvariant inv : tinvs) {
            isKnockedOut = false;
            for(NetViewerNode nvNode : nvNodes) {
                if(inv.factor(project.getPetriNet().findTransition(nvNode.getId())) > 0) {
                    isKnockedOut = true;
                    break;
                }
            }
            if(!isKnockedOut)
                notKnockedOutTinvariants.add(inv);
        }

        for(Transition t : project.getPetriNet().transitions())
            factors.put(t, 0);

        for(TInvariant inv : notKnockedOutTinvariants) {
            for(Transition t : inv) {
                factor = inv.factor(t);
                if(factor > 0) {
                    if(!notKnockedOutTransitions.contains(t))
                        notKnockedOutTransitions.add(t);

                    if(mode == 1)
                        factors.put(t, factors.get(t)+1);
                    else if(mode == 2)
                        factors.put(t, factors.get(t)+factor);
                }
            }
        }

        min = (float)Collections.min(factors.values());
        max = (float)Collections.max(factors.values());
        maxmin = max-min;

        percentTransitions = Math.round(100.0 - Math.round( (((double)notKnockedOutTransitions.size() / (double)project.getPetriNet().transitions().size()) *100) *100)/100.0);
        percentInvariants = Math.round(100.0 - Math.round( ((double)notKnockedOutTinvariants.size() / (double)tinvs.size()) *100 *100)/100.0);

        resetColor();
        
        for(Transition t : project.getPetriNet().transitions()) {
            if(notKnockedOutTransitions.contains(t)) {
                if(mode != 0) {
                    norm = (factors.get(t) - min) / maxmin;
                    if(norm < 0.05)
                        norm = (float)0.05;
                    transitionMap.get(t.id()).setColor(new Color(Color.HSBtoRGB(hsbvals[0], norm, hsbvals[2])));
                }
                else
                   transitionMap.get(t.id()).setColor(notKnockedOutColor);
            }
            else
                transitionMap.get(t.id()).setColor(alsoKnockedOutColor);
        }

        String knockOutString = strings.get("NVKnockOutMessage", percentInvariants, percentTransitions);
        String heatMapString = (mode != 0) ? "<br>" + strings.get("NVImportantTransitionMessage", (int)min, (int)max) : "";
        messageLabel.setText("<html>"+knockOutString +"<br />"+ heatMapString+"</html>");

        for(NetViewerNode nvNode : nvNodes)
            transitionMap.get(nvNode.getId()).setColor(knockedOutColor);
        
        vv.repaint();
    }

    protected void resetKnockOut() {
        resetColor();
        resetMessageLabel();
    }

    /**
     * Set the messageLabel to empty text and set color to black
     */
    private void resetMessageLabel() {
        messageLabel.setText("");
        messageLabel.setForeground(Color.BLACK);
    }

    /**
     * Set the bessageLabel to given text and color
     * @param text
     * @param color
     */
    protected void setMessageLabelTo(String text, Color color) {
        messageLabel.setText(text);
        messageLabel.setForeground(color);
    }

    /**
     * Action on adding a new inVertex
     */
    protected void inVertexMouseAction() {
        changeMouseModeToPicking();
        vv.getPickedVertexState().clear();
        markSelectedMouseMode(tbf.inEdgePanel);
        setMessageLabelTo(strings.get("NVInsertInVertexMessage"), Color.BLACK);
        gpmp.setMouseModeToInVertex();
    }

    /**
     * Action on adding a new outVertex
     */
    protected void outVertexMouseAction() {
        changeMouseModeToPicking();
        vv.getPickedVertexState().clear();
        markSelectedMouseMode(tbf.outEdgePanel);
        setMessageLabelTo(strings.get("NVInsertOutVertexMessage"), Color.BLACK);
        gpmp.setMouseModeToOutVertex();
    }

    /**
     * Action on adding a new edge
     */
    protected void edgeMouseAction() {
        vv.getPickedVertexState().clear();  
        // Adding only one edge
        if(gpmp.getMouseMode().equalsIgnoreCase(GraphPopupMousePlugin.SINGLE_EDGE)) {
            cancelMouseAction();
        }
        // or more edges?
        else {
            changeMouseModeToPicking();
            markSelectedMouseMode(tbf.addEdgePanel);
            setMessageLabelTo(strings.get("NVInsertDoubleEdgeMessage"), Color.BLACK);
            gpmp.setMouseModeToDoubleEdge();
        }              
    }

    /**
     * Action on adding a new place
     */
    protected void placeMouseAction() {
        changeMouseModeToPicking();
        vv.getPickedVertexState().clear();
        mml.setMouseModeToPlace();
        markSelectedMouseMode(tbf.addPlacePanel);
        setMessageLabelTo(strings.get("NVPlaceMessage"), Color.BLACK);
    }

   /**
    * Action on adding a new transition
    */
   protected void transitionMouseAction() {
        changeMouseModeToPicking();
        vv.getPickedVertexState().clear();
        mml.setMouseModeToTransition();
        markSelectedMouseMode(tbf.addTransitionPanel);
        setMessageLabelTo(strings.get("NVTransitionMessage"), Color.BLACK);
    }

    /**
     * Action on remove an edge
     */
    protected void deleteMouseAction() {
        changeMouseModeToPicking();
        vv.getPickedEdgeState().clear();
        vv.getPickedVertexState().clear();
        gpmp.setMouseModeToDelete();
        markSelectedMouseMode(tbf.deletePanel);
        setMessageLabelTo(strings.get("NVDeleteMessage"), Color.BLACK);
    }

    /**
     * Action on adding a bend to an edge
     */
    protected void addBendMouseAction() {
        changeMouseModeToPicking();
        vv.getPickedEdgeState().clear();
        gpmp.setMouseModeToAddBend();
        markSelectedMouseMode(tbf.addBendPanel);
        setMessageLabelTo(strings.get("NVAddBendMessage"), Color.BLACK);
    }

    /**
     * Action on removing a bend from an edge
     */
    protected void deleteBendMouseAction() {
        changeMouseModeToPicking();
        vv.getPickedEdgeState().clear();
        gpmp.setMouseModeToDeleteBend();
        markSelectedMouseMode(tbf.removeBendPanel);
        setMessageLabelTo(strings.get("NVDeleteBendMessage"), Color.BLACK);
    }

    protected void alignXMouseAction() {
        changeMouseModeToPicking();
        gpmp.setMouseModeToAlignX();
        markSelectedMouseMode(tbf.allignXPanel);
        saveSelectedVertices();
        setMessageLabelTo(strings.get("NVAlignmentText"), Color.BLACK);
    }

    protected void alignYMouseAction() {
        changeMouseModeToPicking();
        gpmp.setMouseModeToAlignY();
        markSelectedMouseMode(tbf.allignYPanel);
        saveSelectedVertices();
        setMessageLabelTo(strings.get("NVAlignmentText"), Color.BLACK);
    }

    protected void exportSelectedSubGraphMouseAction() throws FileNotFoundException {
        JFileChooser petriNetFileChooser = new JFileChooser(project.getPath());
        petriNetFileChooser.setDialogType(JFileChooser.SAVE_DIALOG);
        petriNetFileChooser.setDialogTitle(strings.get("ChooseAPetriNetFileName"));
        petriNetFileChooser.setAcceptAllFileFilterUsed(false);

        for (final OutputHandler handler : PetriNetOutputHandlers.getHandlers()) {
            petriNetFileChooser.addChoosableFileFilter(new MyOutputFileFilter(handler));
        }

        if (petriNetFileChooser.showDialog(this, "Export") != JFileChooser.APPROVE_OPTION)
            return;
        
        PetriNet subNet = new PetriNet();
        Set<NetViewerEdge> possibleEdges = new HashSet<NetViewerEdge>();
        
        for(NetViewerNode n : vv.getRenderContext().getPickedVertexState().getPicked()) {
            if(n.getNodeType().equalsIgnoreCase(NetViewer.PLACE)) {
                subNet.addPlace(project.getPetriNet().findPlace(n.getMasterNode().getId()));
            } 
            else if(n.getNodeType().equalsIgnoreCase(NetViewer.TRANSITION)) {
                subNet.addTransition(project.getPetriNet().findTransition(n.getMasterNode().getId()));
            }
            
            for(NetViewerEdge e : n.getInEdges()) {
                possibleEdges.add(e);
            }            
            for(NetViewerEdge e : n.getOutEdges()) {
                possibleEdges.add(e);
            }            
        }
        
        for(NetViewerEdge e : possibleEdges) {            
            if(e.getSource().getNodeType().equalsIgnoreCase(NetViewer.PLACE)) {   
                if(subNet.findPlace(e.getSource().getId()) != null && subNet.findTransition(e.getAim().getId()) != null) {
                    subNet.addArc(subNet.findPlace(e.getSource().getId()), subNet.findTransition(e.getAim().getId()), e.getWeight());
                }
            }  
            else if(e.getSource().getNodeType().equalsIgnoreCase(NetViewer.TRANSITION)) {                
                 if(subNet.findTransition(e.getSource().getId()) != null && subNet.findPlace(e.getAim().getId()) != null) {
                     subNet.addArc(subNet.findTransition(e.getSource().getId()), subNet.findPlace(e.getAim().getId()), e.getWeight());
                 }
            }           
        }           
        
        File petriNetFile = petriNetFileChooser.getSelectedFile();
        MyOutputFileFilter selectedFileFilter = ((MyOutputFileFilter)petriNetFileChooser.getFileFilter());
        petriNetFile = selectedFileFilter.checkFileNameForExtension(petriNetFile);
        selectedFileFilter.getHandler().save(new FileOutputStream(petriNetFile), subNet);        
    }    
    
    /**
     * Cancle mouse action and set mouse to picking mode
     */
    protected void cancelMouseAction() {
        gpmp.setMouseModeToNormal();
        mml.setMouseModeToNormal();
        markSelectedMouseMode(tbf.mousePickingPanel);
        resetMessageLabel();
        netChanged();
    }

    /**
     * Marks the button of the current MouseMode
     * @param activatedButton
     */
    private void markSelectedMouseMode(JPanel activatedPanel) {
        for(JPanel p : mouseMeodePanels) {
            if(p.equals(activatedPanel)) {  
                p.setBackground(Color.RED);                                
            }
            else {             
                p.setBackground(defaultPanelColor);                
            }
        }
    }

    private void saveSelectedVertices() {
        if(vv.getRenderContext().getPickedVertexState().getPicked().isEmpty())
            return;
        alignmentList.clear();
        for(NetViewerNode nvNode : vv.getRenderContext().getPickedVertexState().getPicked())
            alignmentList.add(nvNode);

        vv.getRenderContext().getPickedVertexState().clear();
    }

    protected void alignVertices(String pos, NetViewerNode nodeToAlign) {
        int newX = (int) layout.transform(nodeToAlign).getX();
        int newY = (int) layout.transform(nodeToAlign).getY();
        
        if(pos.equalsIgnoreCase("X")) {
            for(NetViewerNode nvNode : alignmentList)
                layout.setLocation(nvNode, new Point(newX, (int) layout.transform(nvNode).getY()));
        }
        else if(pos.equalsIgnoreCase("Y")) {
            for(NetViewerNode nvNode : alignmentList)
                layout.setLocation(nvNode, new Point((int) layout.transform(nvNode).getX(), newY));
        }
        cancelMouseAction();
        resetMessageLabel();
        vv.repaint();
    }

    protected void deleteSelectedVertices() {
        cancelMouseAction();
        List<NetViewerNode> selectedNodes = new ArrayList<NetViewerNode>();

        for(NetViewerNode nvNode : vv.getRenderContext().getPickedVertexState().getPicked())
            selectedNodes.add(nvNode);

        if(!selectedNodes.isEmpty()) {
            for(NetViewerNode nvNode : selectedNodes) {
                if(!nvNode.getNodeType().equalsIgnoreCase(NetViewer.BEND))
                    deleteNode(nvNode, false);
            }
        }

        List<NetViewerEdge> selectedEdges = new ArrayList<NetViewerEdge>();

        for(NetViewerEdge nvEdge : vv.getRenderContext().getPickedEdgeState().getPicked())
            selectedEdges.add(nvEdge);

        if(!selectedEdges.isEmpty()) {
            for(NetViewerEdge nvEdge : selectedEdges)
                deleteEdge(nvEdge.getMasterEdge(), false);
        }

        updateInfoBar();
    }


    /**
     * Creates a Petri net from the current VisualizationViewer
     */
    private PetriNet createPetriNet() {
        PetriNet pn = new PetriNet();

        Place place;
        Transition transition;
        String propertyName;
        for(NetViewerNode nvNode : g.getVertices()) {
            if(nvNode.getNodeType().equalsIgnoreCase(NetViewer.PLACE)) {
                if(nvNode.isMasterNode() || (!nvNode.isMasterNode() && !nvNode.isLogical()) ) {
                    place = new Place(nvNode.getId());
                    if(place == null) {
                        continue;
                    }
                    place.putProperty("name", nvNode.getName());
                    place.putProperty("posX", layout.transform(nvNode).getX());
                    place.putProperty("posY", layout.transform(nvNode).getY());
                    
                    Iterator it = nvNode.getPropertyList().iterator();
                    while(it.hasNext()) {
                        propertyName = (String) it.next();
                        if(!place.hasProperty(propertyName))
                            place.putProperty(propertyName, nvNode.getProperty(propertyName));
                    }
                    
                    pn.addPlace(place);
                }
            } else if(nvNode.getNodeType().equalsIgnoreCase(NetViewer.TRANSITION)) {
                transition = new Transition(nvNode.getId());
                if(transition == null) {
                    continue;
                }                
                transition.putProperty("name", nvNode.getName());
                transition.putProperty("posX", layout.transform(nvNode).getX());
                transition.putProperty("posY", layout.transform(nvNode).getY());
                
                Iterator it = nvNode.getPropertyList().iterator();
                while(it.hasNext()) {
                    propertyName = (String) it.next();
                    if(!transition.hasProperty(propertyName))
                        transition.putProperty(propertyName, nvNode.getProperty(propertyName));
                }
                
                pn.addTransition(transition);
            }
        }

        String aimType, sourceType;
        Arc arc = null;
        for(NetViewerEdge nvEdge : g.getEdges()) {
            aimType = nvEdge.getAim().getNodeType();
            sourceType = nvEdge.getSource().getNodeType();

            if(!aimType.equalsIgnoreCase(NetViewer.BEND) && !sourceType.equalsIgnoreCase(NetViewer.BEND)) {
                if(sourceType.equalsIgnoreCase(NetViewer.PLACE)) {
                    arc = new Arc(pn.findPlace(nvEdge.getSource().getMasterNode().getId()), pn.findTransition(nvEdge.getAim().getId()), nvEdge.getWeight());
                    if(arc.source() == null || arc.aim() == null)
                        continue;                    
                    pn.addArc((Place)arc.source(), (Transition)arc.aim(), arc);
                }
                else if(sourceType.equalsIgnoreCase(NetViewer.TRANSITION)) {
                    arc = new Arc(pn.findTransition(nvEdge.getSource().getId()), pn.findPlace(nvEdge.getAim().getMasterNode().getId()), nvEdge.getWeight());
                    if(arc.source() == null || arc.aim() == null)
                        continue;                    
                    pn.addArc((Transition)arc.source(), (Place)arc.aim(), arc);
                }
            }
        }

        return pn;
    }

    /**
     * Update all properties of the Petri net with new names ect.
     * @throws IOException
     * @throws ClassNotFoundException
     */
    public void updatePetriNet() throws IOException, ClassNotFoundException {
        updatePetriNet(true);
    }

    /**
     * Update all properties of the Petri net with new names ect.
     * @param saveLayout
     * @throws IOException
     * @throws ClassNotFoundException
     */
    public void updatePetriNet(Boolean saveLayout) throws IOException, ClassNotFoundException {
        PetriNet pn = this.project.getPetriNet();
        NetViewerNode nvNode = null;
        for(Transition t : pn.transitions()) {
            nvNode = transitionMap.get(t.id());
            t.putProperty("name", nvNode.getName());
            t.putProperty("posX", layout.transform(nvNode).getX());
            t.putProperty("posY", layout.transform(nvNode).getY());
        }
        for(Place p : pn.places()) {
            nvNode = placeMap.get(p.id());
            p.putProperty("name", nvNode.getName());
            p.putProperty("posX", layout.transform(nvNode).getX());
            p.putProperty("posY", layout.transform(nvNode).getY());
        }
        
        String aimType, sourceType;
        Arc arc = null;
        for(NetViewerEdge nvEdge : g.getEdges()) {
            aimType = nvEdge.getAim().getNodeType();
            sourceType = nvEdge.getSource().getNodeType();

            if(!aimType.equalsIgnoreCase(NetViewer.BEND) && !sourceType.equalsIgnoreCase(NetViewer.BEND)) {
                if(sourceType.equalsIgnoreCase(NetViewer.PLACE))
                    arc = pn.getArc(pn.findPlace(nvEdge.getSource().getMasterNode().getId()), pn.findTransition(nvEdge.getAim().getMasterNode().getId()));
                else if(sourceType.equalsIgnoreCase(NetViewer.TRANSITION))
                    arc = pn.getArc(pn.findTransition(nvEdge.getSource().getMasterNode().getId()), pn.findPlace(nvEdge.getAim().getMasterNode().getId()));

                if(arc == null)
                    continue;
                arc.setWeight(nvEdge.getWeight());
            }
        }

        saveView(true, true);
    }

    private void showNodeDegrees() {
        Map<Integer, Integer> ndPlaces = new HashMap<Integer, Integer>();
        Map<Integer, Integer> ndTransitions = new HashMap<Integer, Integer>();

        Integer count = 0;
        for(Place p : project.getPetriNet().places()) {
            count = 0;

            count += p.inputs().size() + p.outputs().size();
            if(!ndPlaces.containsKey(count))
                ndPlaces.put(count, 0);
            ndPlaces.put(count, ndPlaces.get(count)+1);
        }

        List<Integer> listOfPlaceDegrees = new ArrayList<Integer>(ndPlaces.keySet());
        Collections.sort(listOfPlaceDegrees);

        for(Transition t : project.getPetriNet().transitions()) {
            count = 0;

            count += t.inputs().size() + t.outputs().size();
            if(!ndTransitions.containsKey(count))
                ndTransitions.put(count, 0);
            ndTransitions.put(count, ndTransitions.get(count)+1);
        }

        List<Integer> listOfTransitionDegrees = new ArrayList<Integer>(ndTransitions.keySet());
        Collections.sort(listOfTransitionDegrees);
        // Create the Histrogramm
        NodesDegrees nodesDegrees = new NodesDegrees(ndPlaces, ndTransitions, project.getPetriNet().places().size(), project.getPetriNet().transitions().size());
    }

    public void closeAllFrames() {
        for(JFrame frame : framesList) {
            frame.setVisible(false);
            frame.dispose();
        }
    }

    /**
     * Shows the NetViewer
     */
    public void showMe() {
        this.setVisible(true);
    }

    /**
     * Changes the color of given label
     * @param label
     * @param color
     */
    private void changeLabelColor(JLabel label, Color color) {
        label.setForeground(color);
        label.setBackground(color);
    }

     /**
     * Shows the properties menu for a vertex
     * @param nvNode
     * @param x
     * @param y
     */
    protected void showVertexSetup(NetViewerNode nvNode, int x, int y) {
        // Caution is the mother of wisdom
        if(!nvNode.getNodeType().equalsIgnoreCase(NetViewer.BEND)) {
            vsf.setLocation(x, y);

            vsf.vertexTypeLabel.setEnabled(true);
            vsf.circleRadioButon.setEnabled(true);
            vsf.polygoneRadioButton.setEnabled(true);
            vsf.cornerSpinner.setEnabled(true);
            vsf.vertexNameTextField.setEnabled(true);
            vsf.vertexNoteTextArea.setEnabled(true);            
            
            vsf.vertexNameLabel.setText(nvNode.getName());
            vsf.vertexTypeLabel.setText("["+nvNode.getReadableNodeType()+"]");
            if(nvNode.getCorners() == 0)
                vsf.circleRadioButon.setSelected(true);
            else
                vsf.polygoneRadioButton.setSelected(true);
            vsf.cornerSpinner.setValue((nvNode.getCorners() == 0) ? 4 : nvNode.getCorners());
            vsf.showColorLabelVertex.setForeground(nvNode.getColor());
            vsf.showColorLabelVertex.setBackground(nvNode.getColor());
            vsf.vertexNameTextField.setText(nvNode.getName());
            if(nvNode.hasProperty("toolTip"))
                vsf.vertexNoteTextArea.setText((String) nvNode.getProperty("toolTip"));
            else
                vsf.vertexNoteTextArea.setText("");

            Position position = nvNode.getLabelPosition();

            if(position == null)
                vsf.SE.setSelected(true);
            else {
                if(position.equals(Position.CNTR))
                    vsf.C.setSelected(true);
                else if(position.equals(Position.E))
                    vsf.E.setSelected(true);
                else if(position.equals(Position.N))
                    vsf.N.setSelected(true);
                else if(position.equals(Position.NE))
                    vsf.NE.setSelected(true);
                else if(position.equals(Position.NW))
                    vsf.NW.setSelected(true);
                else if(position.equals(Position.S))
                    vsf.S.setSelected(true);
                else if(position.equals(Position.SE))
                    vsf.SE.setSelected(true);
                else if(position.equals(Position.SW))
                    vsf.SW.setSelected(true);
                else if(position.equals(Position.W))
                    vsf.W.setSelected(true);
            }

            vsf.setVisible(true);
            vsf.setAlwaysOnTop(true);
            this.setEnabled(false);
        }
    }

     /**
     * Shows the properties menu for a set of vertices
     * @param nvNode
     * @param x
     * @param y
     */
    protected void showVertexSetup(List<NetViewerNode> selectedNodes, int x, int y) {
        // Caution is the mother of wisdom
        if(selectedNodes != null) {
            selection4Modification = selectedNodes;
            vsf.setLocation(x, y);

            vsf.vertexNameLabel.setText("Multi selection");
            vsf.vertexTypeLabel.setText("");
            vsf.vertexTypeLabel.setEnabled(false);
            vsf.circleRadioButon.setEnabled(false);
            vsf.circleRadioButon.setSelected(false);
            vsf.polygoneRadioButton.setEnabled(false);
            vsf.polygoneRadioButton.setSelected(false);
            vsf.cornerSpinner.setEnabled(false);
            vsf.showColorLabelVertex.setForeground(Color.BLACK);
            vsf.showColorLabelVertex.setBackground(Color.BLACK);
            vsf.vertexNameTextField.setText("");
            vsf.vertexNameTextField.setEnabled(false);
            vsf.vertexNoteTextArea.setText("");
            vsf.vertexNoteTextArea.setEnabled(false);

            vsf.setVisible(true);
            vsf.setAlwaysOnTop(true);
            this.setEnabled(false);
        }
    }

    /**
     * Shows the properties menu of a edge
     * @param nvEdge
     * @param x
     * @param y
     */
    public void showEdgeSetup(NetViewerEdge nvEdge, int x, int y) {
        esf.setLocation(x, y);

        esf.weightSpinner.setValue(nvEdge.getWeight());
        esf.showColorLabelEdge.setForeground(nvEdge.getColor());
        esf.showColorLabelEdge.setBackground(nvEdge.getColor());
        if(nvEdge.hasProperty("toolTip"))
            esf.edgeNoteTextArea.setText((String) nvEdge.getProperty("toolTip"));

        esf.setVisible(true);
        esf.setAlwaysOnTop(true);
        this.setEnabled(false);
    }

    /**
     * Shows the logical place menu
     * @param nvNode
     * @param x
     * @param y
     */
    protected void showCreateLogicalPlace(NetViewerNode nvNode, int x, int y) {
        lpf.setLocation(x, y);

        neighborsListModel.removeAllElements();
        for(NetViewerNode n : g.getNeighbors(nvNode)) {
            if(!n.getNodeType().equalsIgnoreCase(NetViewer.BEND)) {
                neighborsListModel.addElement(n);
            }
        }

        lpf.setVisible(true);
        lpf.setAlwaysOnTop(true);
        this.setEnabled(false);
    }

    /**
     * Shows dialog to rename a view
     */
    protected void showEditView() {
        String name = JOptionPane.showInputDialog(strings.get("NVRenameView"), ((View)tbf.viewsCb.getSelectedItem()).toString());
        if(name != null)
            ((View)tbf.viewsCb.getSelectedItem()).setName(name);
        tbf.viewsCb.repaint();
    }

    /**
     * Show a dialog to change the color settings
     */
    private void showColorOptions() {
        changeLabelColor(cof.invColorSelectionLabel, Settings.getAsColor("tinvColor"));
        changeLabelColor(cof.mctsColorSelectionLabel, Settings.getAsColor("mctsColor"));
        changeLabelColor(cof.heatMapColorSelectionLabel, Settings.getAsColor("heatMapColor"));
        changeLabelColor(cof.knockedOutColorSelectionLabel, Settings.getAsColor("knockedOutColor"));
        changeLabelColor(cof.alsoKnockedOutColorSelectionLabel, Settings.getAsColor("alsoKnockedOutColor"));
        changeLabelColor(cof.notKnockedOutColorSelectionLabel, Settings.getAsColor("notKnockedOutColor"));
        changeLabelColor(cof.backgroundColorSelectionLabel, Settings.getAsColor("backgroundColor"));

        cof.setEnabled(true);
        cof.setVisible(true);
        cof.setAlwaysOnTop(true);
        this.setEnabled(false);
    }
    
    private void exitNetViewer() {
        this.setVisible(false);
    }

    private void initExternalFrames() {
        // Create ToolBar
        tbf = new ToolBarFrame(this, nvkl, colorMap);

        // Invariants ComboBoxes
        tinvCbList.add(tbf.allInvCb);
        tinvCbList.add(tbf.trivialInvCb);
        tinvCbList.add(tbf.ioInvCb);
        tinvCbList.add(tbf.inputInvCb);
        tinvCbList.add(tbf.outputInvCb);
        tinvCbList.add(tbf.cyclicInvCb);

        // Fill the Boxes
        // Are T-imvariants available? If, than add these to the ComboBoxes
        tinvs = getTInvs();
        if(tinvs == null) {
            for(JComboBox cb : tinvCbList)
                cb.setEnabled(false);
        }
        else
            addTinvsToComboBox(tinvs);

        // Are MCTS available? If, than add these to the ComboBox
        if(mctsResults == null) {
            tbf.mctsCb.setEnabled(false);
        }
        else
            addMctsToComboBox();

        // Are P-imvariants available? If, than add these to the ComboBox
        pinvs = getPInvs();
        if(pinvs == null)
            tbf.pinvCb.setEnabled(false);
        else
            addPinvsToComboBox();

        // Add ItemListener for load a  View
        vpil = new ViewsPopupItemListener();
        tbf.viewsCb.addItemListener(vpil);

        // Add all MouseMode Buttons to a list
        mouseMeodePanels.add(tbf.mousePickingPanel);
        mouseMeodePanels.add(tbf.mouseTransformingPanel);
        mouseMeodePanels.add(tbf.inEdgePanel);
        mouseMeodePanels.add(tbf.outEdgePanel);
        mouseMeodePanels.add(tbf.addEdgePanel);
        mouseMeodePanels.add(tbf.deletePanel);
        mouseMeodePanels.add(tbf.addBendPanel);
        mouseMeodePanels.add(tbf.removeBendPanel);
        mouseMeodePanels.add(tbf.addPlacePanel);
        mouseMeodePanels.add(tbf.addTransitionPanel);
        mouseMeodePanels.add(tbf.allignXPanel);
        mouseMeodePanels.add(tbf.allignYPanel);

        // Create SearchBar
        sb = new SearchBar(this);
        addTabToMenuBar("SearchBar", sb.getContentPane());        
        
        allPlacesModel = (DefaultListModel) sb.allPlacesList.getModel();
        allTransitionsModel = (DefaultListModel) sb.allTransitionsList.getModel();

        addWindowListener(new NetViewerWindowsListener(sb, tbf));

        // Create VertexSetupFrame
        vsf = new VertexSetupFrame(this);

        // create EdgeSetupFrame
        esf = new EdgeSetupFrame(this);

        // create LocigalPlaceFrame
        lpf = new LogicalPlacesFrame(this);
        neighborsListModel = (DefaultListModel) lpf.neighborsList.getModel();

        // create color options frame
        cof = new ColorOptionsFrame(this);

        framesList.add(vsf);
        framesList.add(esf);
        framesList.add(sb);
        framesList.add(tbf);
        framesList.add(cof);
        framesList.add(lpf);
    }

    private void initMenuBar() {
        menuBar = new JMenuBar();

        newViewItem = new JMenuItem(strings.get("NVSaveLayoutMenu"), resources.getIcon("new_project.png"));
        newViewItem.setActionCommand(NEW_VIEW);
        newViewItem.addActionListener(this);
        newViewItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_N, java.awt.event.InputEvent.CTRL_MASK));

        saveViewItem = new JMenuItem(strings.get("NVSaveAsLayoutMenu"), resources.getIcon("save_project.png"));
        saveViewItem.setActionCommand(SAVE_VIEW);
        saveViewItem.addActionListener(this);
        saveViewItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_S, java.awt.event.InputEvent.CTRL_MASK));

        saveAsPictureItem = new JMenuItem(strings.get("NVMakePicButtonShort"), resources.getIcon("save_picture.png"));
        saveAsPictureItem.setActionCommand(MAKE_PIC);
        saveAsPictureItem.addActionListener(this);
        saveAsPictureItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_I, java.awt.event.InputEvent.CTRL_MASK));

        exitItem = new JMenuItem(strings.get("NVClose"), resources.getIcon("delete.png"));
        exitItem.setActionCommand(EXIT);
        exitItem.addActionListener(this);
        exitItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_X, java.awt.event.InputEvent.CTRL_MASK));

        fileMenu = new JMenu(strings.get("File"));
        fileMenu.setMnemonic(KeyEvent.VK_F);
        fileMenu.add(newViewItem);
        fileMenu.add(saveViewItem);
        fileMenu.add(saveAsPictureItem);
        fileMenu.addSeparator();
        fileMenu.add(exitItem);

        viewMenu = new JMenu(strings.get("NVViewMenu"));
        viewMenu.setMnemonic(KeyEvent.VK_V);

        colorItem = new JMenuItem(strings.get("NVHideColorText"));
        colorItem.setActionCommand(HIDE_COLOR);
        colorItem.addActionListener(this);

        labelItem = new JMenuItem(strings.get("NVHideLabel"));
        labelItem.setActionCommand(SHOW_LABELS);
        labelItem.addActionListener(this);

        centerItem = new JMenuItem(strings.get("NVCenterPetriNet"));
        centerItem.setActionCommand(CENTER_NET);
        centerItem.addActionListener(this);

        resetColoringItem = new JMenuItem(strings.get("NVResetColoring"));
        resetColoringItem.setActionCommand(RESET_COLORING);
        resetColoringItem.addActionListener(this);        
        
        ctiItem = new JMenuItem(strings.get("NVCTITransitionen"));
        ctiItem.setActionCommand(CTI_TRANSITIONS);
        ctiItem.addActionListener(this);

        byOccurrenceItem = new JMenuItem(strings.get("NVByOccurrence"));
        byOccurrenceItem.setActionCommand(BY_OCCURRENCE);
        byOccurrenceItem.addActionListener(this);

        byFactorItem = new JMenuItem(strings.get("NVByFactor"));
        byFactorItem.setActionCommand(BY_FACTOR);
        byFactorItem.addActionListener(this);

        importantTransitonsMenu = new JMenu(strings.get("NVImportantTransitions"));
        importantTransitonsMenu.add(byOccurrenceItem);
        importantTransitonsMenu.add(byFactorItem);

        visualizationMenu = new JMenu(strings.get("NVVisualization"));     
        visualizationMenu.setMnemonic(KeyEvent.VK_I);
        visualizationMenu.add(colorItem);
        visualizationMenu.add(labelItem);
        visualizationMenu.add(centerItem);
        visualizationMenu.add(resetColoringItem);
        visualizationMenu.add(ctiItem);
        visualizationMenu.add(importantTransitonsMenu);

        nodeDegreeItem = new JMenuItem(strings.get("NVNodeDegree"));
        nodeDegreeItem.setActionCommand(NODE_DEGREE);
        nodeDegreeItem.addActionListener(this);

        topolocigalMenu = new JMenu(strings.get("NVTopological"));
        topolocigalMenu.add(nodeDegreeItem);

        colorOptionsItem = new JMenuItem(strings.get("NVColorOptions"));
        colorOptionsItem.setActionCommand(SHOW_COLOR_OPTION);
        colorOptionsItem.addActionListener(this);

        optionsMenu = new JMenu(strings.get("NVOptions"));
        optionsMenu.setMnemonic(KeyEvent.VK_O);
        optionsMenu.add(colorOptionsItem);

        menuBar.add(fileMenu);
//        menuBar.add(viewMenu);
        menuBar.add(visualizationMenu);
        menuBar.add(topolocigalMenu);
        menuBar.add(optionsMenu);
        setJMenuBar(menuBar);
    }

    // ------- START: Communication and controll with / for other addons -----------

    /**
     * Add a new Tab to the Menu Bar of the NetViewer
     * @param name The name of the tab, shown in the header of the tab
     * @param tab The Component which is shown in the Tab. (Panel or ToolBar). Please use a TableLayout.
     */
    public void addTabToMenuBar(String name, Component tab) {
        tbf.addTabToMenuBar(name, tab);
    }

    /**
     * Returns the actual VisualizationViewer
     * @return VisualizationViewer
     */
    public VisualizationViewer getVisualizationViewer() {
        return vv;
    }

    /**
     * Return the actual vertex size
     * @return vertexSize
     */
    public int getVertexSize() {
        return vertexSize;
    }

    /**
     * This map connects the id of a place in the Petri net with the corresponding NetViewerNode
     * @return
     */
    public Map<Integer, NetViewerNode> getPlaceMap() {
        return placeMap;
    }

    /**
     * This map connects the id of a transition in the Petri net with the corresponding NetViewerNode
     * @return
     */
    public Map<Integer, NetViewerNode> getTransitionMap() {
        return transitionMap;
    }

    // ------- END: Communication and controll with / for other addons -----------

    /*
    **********************Internal classes**************************************
    */

    /**
     *  ComboBox PopupListener for coloring T-invariants or MCT-sets
     */
    static protected class TinvItemListener implements ItemListener {
        private JComboBox cb;
        private Boolean colorPlaces;
        private Color color;
        private NetViewer nv;

       public TinvItemListener(NetViewer nv, JComboBox cb, Boolean colorPlaces, Color color) {
            this.cb = cb;
            this.colorPlaces = colorPlaces;
            this.color = color;
            this.nv = nv;
        }

        public void itemStateChanged(ItemEvent ie) {
            if(ie.getStateChange() == ItemEvent.SELECTED) {
                // Nothing is selected
                if(cb.getSelectedItem() == null)
                    return;
                if(cb.getSelectedIndex() == 0) {
                    nv.resetColor();
                    return;
                }
                // If, for what ever reason, the item is not a Wrapper
                if(!cb.getSelectedItem().getClass().getSimpleName().equalsIgnoreCase("TinvWrapper"))
                    return;
                
                nv.resetMessageLabel();
                
                if(!nv.tbf.stackTinvs.isSelected()) {
                    nv.resetColor();                    
                } 
                
                Color choosenColor;
                if(nv.tbf.manuellColorSelection.isSelected()) {
                    choosenColor = JColorChooser.showDialog(null, "Select color", null);
                } else {
                    choosenColor = this.color;
                }             
                                
                Set<Transition> transitions;  
                Iterator<Transition> it;
                Transition t;                
                float min = 0, max, maxmin = 0;
                float[] hsbvals = null;
                float norm;
                List<Integer> allFactors = new ArrayList<Integer>();                
                    
                TInvariant tinv = ((TinvWrapper)cb.getSelectedItem()).getTinv();
                transitions = new HashSet<Transition>(tinv.transitions());

                // Find minimum and maximum value
                if(nv.heatMap) {
                    min = 0;
                    maxmin = 0;
                    hsbvals = Color.RGBtoHSB(choosenColor.getRed(), choosenColor.getGreen(), choosenColor.getBlue(), null);
                    allFactors.clear();                                                                                        
                    it = transitions.iterator();
                    while(it.hasNext()) {
                        allFactors.add(tinv.factor(it.next()));
                    }
                    min = (float)Collections.min(allFactors);
                    max = (float)Collections.max(allFactors);
                    maxmin = max-min;
                }

                // Coloring of the transitions, edges and places                   
                it = transitions.iterator();                                       
                while(it.hasNext()) {
                    t = it.next();
                    if(nv.heatMap) {
                        norm = (tinv.factor(t) - min) / maxmin;
                        if(norm < 0.05)
                            norm = (float)0.05;
                        choosenColor = new Color(Color.HSBtoRGB(hsbvals[0], norm, hsbvals[2]));
                    }

                    nv.transitionMap.get(t.id()).setColor(choosenColor);
                    if(colorPlaces) {
                        // Color all in/out edges and their source/aim places
                        for(NetViewerEdge nvEdge : nv.transitionMap.get(t.id()).getOutEdges()) {
                            nvEdge.setColorForAllEdges(choosenColor);
                            if(nvEdge.getAim().getNodeType().equals(NetViewer.PLACE))
                                nvEdge.getAim().setColorForAllNodes(choosenColor);
                        }

                        for(NetViewerEdge nvEdge : nv.transitionMap.get(t.id()).getInEdges()) {
                            nvEdge.setColorForAllEdges(choosenColor);
                            if(nvEdge.getSource().getNodeType().equals(NetViewer.PLACE))
                                nvEdge.getSource().setColorForAllNodes(choosenColor);
                        }
                    }
                }
                
                
                nv.vv.repaint();
            }
        }
    }

     /**
     *  ComboBox PopupListener for coloring P-invariants
     */
    static protected class PinvItemListener implements ItemListener {
        private JComboBox cb;
        private Color color;
        private NetViewer nv;

        /**
         * Init a new Listener
         * @param nv
         * @param cb
         * @param color         *
         */
        public PinvItemListener(NetViewer nv, JComboBox cb, Color color) {
            this.cb = cb;
            this.color = color;
            this.nv = nv;
        }

        public void itemStateChanged(ItemEvent ie) {
            if(ie.getStateChange() == ItemEvent.SELECTED) {
                if(cb.getSelectedItem() == null)
                    return;
                if(cb.getSelectedIndex() == 0) {
                    nv.resetColor();
                    nv.vv.repaint();
                    return;
                }
                if(!cb.getSelectedItem().getClass().getSimpleName().equalsIgnoreCase("PinvWrapper"))
                    return;
               
                nv.resetMessageLabel();
                
                if(!nv.tbf.stackTinvs.isSelected()) {
                    nv.resetColor();                    
                } 

                Color choosenColor;
                if(nv.tbf.manuellColorSelection.isSelected()) {
                    choosenColor = JColorChooser.showDialog(null, "Select color", null);
                } else {
                    choosenColor = this.color;
                }                 
                
                PInvariant pinv = ((PinvWrapper)cb.getSelectedItem()).getPinv();
                Set<Place> places = pinv.places();

                Iterator<Place> it = places.iterator();
                while(it.hasNext()) {
                   nv.placeMap.get(it.next().id()).getMasterNode().setColorForAllNodes(choosenColor);
                }

                nv.vv.repaint();
            }   
        }
    }

    /**
     *  ComboBox PopupListener for coloring MCTs
     */
    static protected class MctsItemListener implements ItemListener {
        private JComboBox cb;
        private Color color;
        private NetViewer nv;

        /**
         * Init a new Listener
         * @param nv
         * @param cb
         * @param color         *
         */
        public MctsItemListener(NetViewer nv, JComboBox cb, Color color) {
            this.cb = cb;
            this.color = color;
            this.nv = nv;
        }

        public void itemStateChanged(ItemEvent ie) {
            if(ie.getStateChange() == ItemEvent.SELECTED) {
                if(cb.getSelectedItem() == null)
                    return;                
                if(cb.getSelectedIndex() == 0) {
                    nv.resetColor();
                    nv.vv.repaint();
                    return;
                }
                if(!cb.getSelectedItem().getClass().getSimpleName().equalsIgnoreCase("MctsWrapper"))
                    return;

                if(!nv.tbf.stackTinvs.isSelected()) {
                    nv.resetColor();                    
                } 
                
                Color choosenColor;
                if(nv.tbf.manuellColorSelection.isSelected()) {
                    choosenColor = JColorChooser.showDialog(null, "Select color", null);
                } else {
                    choosenColor = this.color;
                }                 
                
                nv.resetMessageLabel();
                
                TInvariant mcts = ((MctsWrapper)cb.getSelectedItem()).getMcts();                 
                Set<Transition> transitions = mcts.transitions();
                Iterator<Transition> it = transitions.iterator();
                
                while(it.hasNext()) {
                   nv.transitionMap.get(it.next().id()).setColor(choosenColor);
                }                
                
                nv.vv.repaint();
            }   
        }        
    }        
    
    protected class ViewsPopupItemListener implements ItemListener {
        
        public void itemStateChanged(ItemEvent ie) {
            if(ie.getStateChange() == ItemEvent.SELECTED && viewsItemListerEnabeld) {
                if(tbf.viewsCb.getSelectedItem() == null)
                    return;

                // If a selected Item not a String
                if(tbf.viewsCb.getSelectedItem().getClass().equals(View.class)) {
                    View view = (View)tbf.viewsCb.getSelectedItem();
                    try {
                        // Restore the saved graph
                        ByteArrayInputStream bais = new ByteArrayInputStream(view.getGraph());
                        g = (Graph<NetViewerNode, NetViewerEdge>) new ObjectInputStream(bais).readObject();

                        // Write out the layout. This way is needed, because i found no other way to restore the layout
                        File tmpFileLayout = File.createTempFile("MLtmp", ".tmp", FileUtils.getTempDir());
                        BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(tmpFileLayout));
                        bos.write(view.getLayout());
                        bos.close();

                        // Restore the saved results and give these to the project
                        if(view.getResults() != null) {
                            bais = new ByteArrayInputStream(view.getResults());
                            Map<Class<? extends Tool>, Map<Configuration, Result> > results = (Map<Class<? extends Tool>, Map<Configuration, Result> >) new ObjectInputStream(bais).readObject();
                            project.setResults(results);
                        }

                        // Set the MousePlugin to the "new" graph (new reference)
                        gpmp.changeGraph(g);
                        newEdgesCounter = g.getEdgeCount()+1; // Reset the EdgeCounter
                        layout.setGraph(g);
                        layout.restore(tmpFileLayout.getAbsolutePath());

                        // Clear and refill the maps with the new references
                        transitionMap.clear();
                        placeMap.clear();
                        name2Node.clear();
                        for(NetViewerNode n : g.getVertices()) {
                            if(n.getNodeType().equalsIgnoreCase(TRANSITION))
                                transitionMap.put(n.getId(), n);
                            else if(n.getNodeType().equalsIgnoreCase(PLACE))
                                placeMap.put(n.getId(), n);
                            name2Node.put(n.getName(), n);
                        }

                        tmpFileLayout.delete();
                        vv.repaint();

                        // Create Petri net from current VisualizationViewer
                        PetriNet newPetriNet = createPetriNet();

                        // Clears all ComboBoxes
                        for(JComboBox cb : tinvCbList) {
                            cb.removeAllItems();
                            cb.setEnabled(false);
                        }
                        tbf.mctsCb.removeAllItems();
                        tbf.mctsCb.setEnabled(false);

                        // Checks if results are available or not
                        Boolean hasResults = false;
                        if(project.hasResults(new TInvariantTool()))
                            hasResults = true;

                        // Has Petri net structure changed?
                        if(!hasResults) {
                            if((newPetriNet.places().size() != project.getPetriNet().places().size())
                                || (newPetriNet.transitions().size() != project.getPetriNet().transitions().size())
                                || (newPetriNet.getInarcs().size() != project.getPetriNet().getInarcs().size())
                                || (newPetriNet.getOutarcs().size() != project.getPetriNet().getOutarcs().size())) {
                                netChanged = true;
                                lastChanged = false;
                                netChanged();
                            }
                            project.changePetriNet(newPetriNet);
                        }
                        else {
                           // Set new Petri net
                            project.changePetriNet(newPetriNet);

                            // Load calculated T-Invariants

                            // Refresh the T-Invariants with the new references of the transition
                            tinvs = getTInvs();
                            if(tinvs != null) {
                                List<TInvariant> listOfTinv = new ArrayList<TInvariant>();
                                Map<Transition, Integer> transitions;
                                Transition newTransitionReference;
                                for(TInvariant tinv : tinvs) {
                                    transitions = new HashMap<Transition, Integer>();
                                    for(Transition transition : tinv) {
                                        newTransitionReference = project.getPetriNet().findTransition(transition.id());
                                        transitions.put(newTransitionReference, tinv.factor(newTransitionReference));
                                    }
                                    listOfTinv.add(new TInvariant(tinv.id(), transitions));
                                }
                                tinvs = new TInvariants(listOfTinv);
                                project.putResult(new TInvariantTool(), new TInvariantsConfiguration(), tinvs);
                                addTinvsToComboBox(tinvs);
                            }

                            // Refresh the P-Invariants with the new references of the places
                            pinvs = getPInvs();
                            if(pinvs != null) {
                                List<PInvariant> listOfPinv = new ArrayList<PInvariant>();
                                Map<Place, Integer> places;
                                Place newPlaceReference;
                                for(PInvariant pinv : pinvs) {
                                    places = new HashMap<Place, Integer>();
                                    for(Place place : pinv) {
                                        newPlaceReference = project.getPetriNet().findPlace(place.id());
                                        places.put(newPlaceReference, pinv.factor(newPlaceReference));
                                    }
                                    listOfPinv.add(new PInvariant(pinv.id(), places));
                                }
                                pinvs = new PInvariants(listOfPinv);
                                project.putResult(new PInvariantTool(), new PInvariantsConfiguration(), pinvs);
                                addPinvsToComboBox();
                            }

                            // Are MCT-Sets always calculated? If, than load them
                            // No reference refreshing needed, because no further calculations based on these
                            if(project.hasResults(new MctsTool())) {
                                mctsResults = null;
                                addMctsToComboBox();
                            }
                        }

                        int latesdId = 0;
                        bendCounter = 0;
                        for(NetViewerNode n : g.getVertices()) {
                            if(n.getId() > latesdId)
                                latesdId = n.getId();
                            if(n.getNodeType().equalsIgnoreCase(NetViewer.BEND))
                                bendCounter++;
                        }
                        vertexCounter = latesdId;

                        updateInfoBar();
                        updateSearchBar(g.getVertices());
                    } catch (ClassNotFoundException ex) {
                        Logger.getLogger(NetViewer.class.getName()).log(Level.SEVERE, null, ex);
                    } catch (IOException ex) {
                        Logger.getLogger(NetViewer.class.getName()).log(Level.SEVERE, null, ex);
                    }
                }                
            }
        }        
    }

    /**
     * Undo Manager
     */
    private class MyUndoManager {

        //private NetViewer netViewer;
        private Stack undoStack;

        public MyUndoManager() {
            this.undoStack = new Stack();
        }

        /**
         * Add a Undo Event
         * @param u
         */
        public void addEvent(UndoEvent u) {
            undoStack.push(u);
        }

        /**
         * Clears the Undo History
         */
        public void clearUndoHistroy() {
            undoStack.removeAllElements();
        }

        /**
         * Undo a event
         */
        public void undo() {
            if(!undoStack.empty())
                ((UndoEvent)undoStack.pop()).undo();
        }
    }

    /**
     * Undo Events
     */
    private class UndoEvent {
        public static final String ADD_LOGICAL_PLACE = "alp";
        public static final String REMOVE_LOGICAL_PLACE = "rlp";
        public static final String SAVE_VERTEX = "sv";
        public static final String SAVE_EDGE = "se";
        public static final String INSERT_BEND = "ib";
        public static final String REMOVE_BEND = "rb";
        public static final String ADD_VERTEX = "av";
        public static final String DELETE_VERTEX = "dv";
        public static final String ADD_EDGE = "ae";
        public static final String DELETE_EDGE = "de";
        public static final String REVERSE_TRANSITION = "rt";
        public static final String MELTING_VETICES = "mv";

        private String eventType;
        private NetViewerNode nvNode;
        private NetViewerEdge nvEdge;
        private Color oldColor;
        private int oldCorners;
        private String oldName;
        private Object[] selectedNodes, selected;
        private int weight;
        private Point2D position;

        public UndoEvent(String eventType, NetViewerNode nvNode) {
            this.eventType = eventType;
            this.nvNode = nvNode;
            this.oldColor = nvNode.getColor();
            this.oldCorners = nvNode.getCorners();
            this.oldName = nvNode.getName();
        }

        public UndoEvent(String eventType, NetViewerNode nvNode, Point2D position) {
            this.eventType = eventType;
            this.nvNode = nvNode;
            this.oldColor = nvNode.getColor();
            this.oldCorners = nvNode.getCorners();
            this.oldName = nvNode.getName();
            this.position = position;
        }

        public UndoEvent(String eventType, Object[] selected) {
            this.eventType = eventType;
            this.selected = selected;
        }

        private UndoEvent(String eventType, NetViewerNode sourceNode, Object[] selectedNodes) {
            this.eventType = eventType;
            this.nvNode = sourceNode;
            this.selectedNodes = selectedNodes;
        }

        private UndoEvent(String eventType, NetViewerEdge nvEdge) {
            this.eventType = eventType;
            this.nvEdge = nvEdge;
        }

        private UndoEvent(String eventType, NetViewerEdge nvEdge, int weight) {
            this.eventType = eventType;
            this.nvEdge = nvEdge;
            this.weight = weight;
        }

        private void undo() {
            if(this.eventType.equals(ADD_LOGICAL_PLACE)) {
                removeLogicalPlace(nvNode, null, true);
            }
            else if(this.eventType.equals(REMOVE_LOGICAL_PLACE)) {
                createLogicalPlace(this.nvNode, this.selectedNodes, true, position);
            }
            else if(this.eventType.equals(SAVE_VERTEX)) {
                vsf.vertexNameTextField.setText(this.oldName);
                vsf.showColorLabelVertex.setForeground(this.oldColor);
                vsf.showColorLabelVertex.setBackground(this.oldColor);
                if(this.oldCorners == 0)
                    vsf.circleRadioButon.setSelected(true);
                else {
                    vsf.polygoneRadioButton.setSelected(true);
                    vsf.cornerSpinner.setValue(this.oldCorners);
                }
                saveVertexSetup(this.nvNode);
            }
            else if(this.eventType.equals(SAVE_EDGE)) {
                this.nvEdge.setWeightForAllEdges(weight);
                vv.repaint();
            }
            else if(this.eventType.equals(INSERT_BEND)) {
                removeBend(this.nvEdge, true);
            }
            else if(this.eventType.equals(REMOVE_BEND)) {
                Point2D point = calcNewPoint(layout.transform(nvEdge.getSource()));
                insertBend(g.findEdge(this.nvEdge.getSource(), this.nvEdge.getAim()), (int) point.getX(), (int) point.getY(), true);
            }
            else if(this.eventType.equals(ADD_VERTEX)) {
                deleteNode(this.nvNode, true);
            }
            else if(this.eventType.equals(DELETE_VERTEX)) {
                g.addVertex(this.nvNode);
                addVertexToPetriNet(this.nvNode);

                for(NetViewerEdge edge : this.nvNode.getInEdges()) {
                    if(edge.getSource() != null) {
                        edge.getSource().addOutEdge(edge);
                        g.addEdge(edge, edge.getSource(), edge.getAim(), EdgeType.DIRECTED);
                        addArcToPetriNet(edge);
                    }
                }

                for(NetViewerEdge edge : this.nvNode.getOutEdges()) {
                    if(edge.getAim() != null) {
                        edge.getAim().addInEdge(edge);
                        g.addEdge(edge, edge.getSource(), edge.getAim(), EdgeType.DIRECTED);
                        addArcToPetriNet(edge);
                    }
                }

                if(this.nvNode.getNodeType().equalsIgnoreCase(NetViewer.PLACE))
                    placeMap.put(this.nvNode.getId(), nvNode);
                else if(this.nvNode.getNodeType().equalsIgnoreCase(NetViewer.TRANSITION))
                    transitionMap.put(this.nvNode.getId(), nvNode);

                vsf.vertexNameTextField.setText(this.oldName);
                vsf.showColorLabelVertex.setForeground(this.oldColor);
                vsf.showColorLabelVertex.setBackground(this.oldColor);
                if(this.oldCorners == 0)
                    vsf.circleRadioButon.setSelected(true);
                else {
                    vsf.polygoneRadioButton.setSelected(true);
                    vsf.cornerSpinner.setValue(this.oldCorners);
                }
                layout.setLocation(nvNode, position);
                saveVertexSetup(this.nvNode);
            }
            else if(this.eventType.equals(DELETE_EDGE)) {
                g.addEdge(this.nvEdge, this.nvEdge.getSource(), this.nvEdge.getAim(), EdgeType.DIRECTED);
                addArcToPetriNet(this.nvEdge);

                this.nvEdge.getSource().addOutEdge(this.nvEdge);
                this.nvEdge.getAim().addInEdge(this.nvEdge);

                vv.repaint();
            }
            else if(this.eventType.equals(ADD_EDGE)) {
                deleteEdge(this.nvEdge, true);
            }
            else if(this.eventType.equals(REVERSE_TRANSITION)) {
                deleteNode(this.nvNode, true);
            }
            else if(this.eventType.equals(MELTING_VETICES)) {
                NetViewerNode n = null, demoltingNode = (NetViewerNode) selected[0];
                List<NetViewerEdge> toDelete = new ArrayList<NetViewerEdge>();

                for(int i = 1; i < selected.length; i++) {
                    n = (NetViewerNode) selected[i];

                    g.addVertex(n);
                    addVertexToPetriNet(n);

                    if(n.getNodeType().equals(NetViewer.TRANSITION))
                        transitionMap.put(n.getId(), n);
                    else
                        placeMap.put(n.getId(), n);

                    for(NetViewerEdge e : n.getOutEdges()) {
                        g.addEdge(e, e.getSource(), e.getAim(), EdgeType.DIRECTED);
                        addArcToPetriNet(e);
                        toDelete.add(g.findEdge(demoltingNode, e.getAim()));
                    }

                    for(NetViewerEdge e : n.getInEdges()) {
                        g.addEdge(e, e.getSource(), e.getAim(), EdgeType.DIRECTED);
                        addArcToPetriNet(e);
                        toDelete.add(g.findEdge(e.getSource(), demoltingNode));
                    }
                }

                for(NetViewerEdge e: toDelete) {
                    g.removeEdge(e);
                    removeArcFromPetriNet(e);
                }

                updateSearchBar(g.getVertices());
                vv.repaint();
            }
        }
    }

}