/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.data.input;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;

import monalisa.data.pn.Arc;
import monalisa.data.pn.PetriNet;
import monalisa.data.pn.Place;
import monalisa.data.pn.Transition;
import monalisa.util.FileUtils;

import javax.xml.stream.XMLStreamException;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;
import org.sbml.jsbml.SBMLReader;
import org.sbml.jsbml.SBMLDocument;
import org.sbml.jsbml.Model;
import org.sbml.jsbml.Reaction;
import org.sbml.jsbml.Species;
import org.sbml.jsbml.SpeciesReference;


/**
 * Input handler for the SBML format.
 * This class parses a SBML-file.
 * Supports SBML all Versions and Levels specified in <a href="http://sbml.org/Documents/Specifications">http://sbml.org/Documents/Specifications</a>
 **/
public final class SbmlInputHandler implements InputHandler {

    private final Map<Integer, Place> places = new HashMap<Integer, Place>();
    private final Map<Integer, Transition> transitions = new HashMap<Integer, Transition>();
    private final Map<String, Integer> species = new HashMap<String, Integer>();

    @Override
    public boolean isKnownFile(File file) throws IOException {
        if ("sbml".equalsIgnoreCase(FileUtils.getExtension(file))) {
            return true;
        }

        if("xml".equalsIgnoreCase(FileUtils.getExtension(file))) {
            SAXBuilder builder = new SAXBuilder();
            Document doc;
            try {
                doc = builder.build(file);
            } catch (Exception e) {
                return false;
            }

            Element root = doc.getRootElement();
            if (!root.getName().equals("sbml")) {
                return false;
            }
            else {
                return true;
            }
        }
        return false;
    }

    @SuppressWarnings("unchecked")
    @Override
    public PetriNet load(InputStream in) throws IOException {
        PetriNet ret = new PetriNet();

        SBMLReader reader = new SBMLReader();
        SBMLDocument doc = null;

        try {
            doc = reader.readSBMLFromStream(in);
        } catch (XMLStreamException ex) {
            Logger.getLogger(SbmlInputHandler.class.getName()).log(Level.ERROR, null, ex);
        }
                
        Model model = doc.getModel();

        Double weight;
        int countPlaces = 0, countTransitions = 0;
        Long tokens;
        String name, id, reactantName, productName, compartmentId;
        Boolean reversible;
        Place place;
        Transition transition, transition_rev = null;

        for(Species s : model.getListOfSpecies()) {
            name = s.getName();
            id = s.getId();
            if(name.equals("")) {
                name = id;
            }            
            compartmentId = s.getCompartment();
            tokens = new Double(s.getInitialAmount()).longValue();
            place = findPlace(countPlaces, ret);
            place.putProperty("name", name);
            place.putProperty("compartment", compartmentId);
            ret.setTokens(place, tokens);
            species.put(id, countPlaces);
            countPlaces++;
        }

        // Transition data section.
        // List of output and input arcs.
        Boolean floatWeighs;
        int multiplikator, subLen, indexOfPoint;
        String weighString;

        for (Reaction r : model.getListOfReactions()) {
            name = r.getName();
            id = r.getId();
            if(name.equals("")) {
                name = id;
            }                
            reversible = r.getReversible();
            floatWeighs = false;
            multiplikator = 1;

            transition = findTransition(countTransitions, ret);
            transition.putProperty("name", name);
            
            if(reversible) {
                countTransitions++;
                transition_rev = findTransition(countTransitions, ret);
                transition_rev.putProperty("name", name+"_rev");                
            }            

            for(SpeciesReference spr : r.getListOfReactants()) {
                weighString = Double.toString(spr.getStoichiometry());
                indexOfPoint = weighString.indexOf(".");
                subLen = weighString.substring(indexOfPoint+1, weighString.length()).length();
                if(subLen > 1 || !weighString.substring(indexOfPoint+1, weighString.length()).equals("0")) {
                    floatWeighs = true;
                    if(subLen > multiplikator) {
                        multiplikator = subLen;
                    }
                }
            }

            for(SpeciesReference spr : r.getListOfProducts()) {
                weighString = Double.toString(spr.getStoichiometry());
                indexOfPoint = weighString.indexOf(".");
                subLen = weighString.substring(indexOfPoint+1, weighString.length()).length();
                if(subLen > 1 || !weighString.substring(indexOfPoint+1, weighString.length()).equals("0")) {
                    floatWeighs = true;
                    if(subLen > multiplikator) {
                        multiplikator = subLen;
                    }
                }
            }

            multiplikator = (int) Math.pow(10.0, new Double(multiplikator));
            Object source, aim;
            // input arcs: (place = reactant) --> transition
            for(SpeciesReference spr : r.getListOfReactants()) {              
         
                reactantName = spr.getSpecies();
                weight = spr.getStoichiometry();

                if(floatWeighs) {
                    weight *= multiplikator;
                }

                source = findPlace(species.get(reactantName), ret);
                aim = transition;

                ret.addArc((Place)source, (Transition)aim, new Arc(source, aim, weight.intValue()));
                
                if(reversible) {                 
                    source = transition_rev;
                    aim = findPlace(species.get(reactantName), ret);
                    ret.addArc((Transition)source, (Place)aim, new Arc(aim, source, weight.intValue()));
                }
            }
            // output arcs: transition --> (place = product)
            for(SpeciesReference spr : r.getListOfProducts()) {
                productName = spr.getSpecies();
                weight = spr.getStoichiometry();

                if(floatWeighs) {
                    weight *= multiplikator;
                }

                source = transition;
                aim = findPlace(species.get(productName), ret);
                
                ret.addArc((Transition)source, (Place)aim, new Arc(source, aim, weight.intValue()));

                if(reversible) {                   
                    source = findPlace(species.get(productName), ret);
                    aim = transition_rev;
                    ret.addArc((Place)source, (Transition)aim, new Arc(aim, source, weight.intValue()));
                }
            }

            countTransitions++;
        }
        
        return ret;
    }
    
    private Place findPlace(int placeId, PetriNet petriNet) {
        Place place = places.get(placeId);
        
        if (place == null) {
            place = new Place(placeId);
            places.put(placeId, place);
            petriNet.addPlace(place);
        }
        return place;
    }

    private Transition findTransition(int transitionId, PetriNet petriNet) {
        Transition transition = transitions.get(transitionId);
        
        if (transition == null) {
            transition = new Transition(transitionId);
            transitions.put(transitionId, transition);
            petriNet.addTransition(transition);
        }
        return transition;
    }
}
