/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.tools.cluster;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.*;

import monalisa.data.Pair;
import monalisa.data.output.DistanceMatrixOutput;
import monalisa.data.pn.TInvariant;
import monalisa.gui.components.ComboBoxItem;
import monalisa.resources.StringResources;
import monalisa.results.ClusterConfiguration;
import monalisa.results.Clustering;
import monalisa.results.Configuration;
import monalisa.results.DistanceMatrix;
import monalisa.results.DistanceMatrixConfiguration;
import monalisa.results.TInvariants;
import monalisa.results.TInvariantsConfiguration;
import monalisa.tools.AbstractTool;
import monalisa.tools.ErrorLog;
import monalisa.tools.Tool;
import monalisa.tools.distance.DistanceFunctions;
import monalisa.tools.distance.DistanceTool;
import monalisa.tools.tinv.TInvariantTool;
import monalisa.util.FileUtils;
import monalisa.Project;
import monalisa.results.ClusterTreeConfiguration;
import monalisa.results.TreeClustering;

public final class ClusterTool extends AbstractTool implements ActionListener {
    private static final String ACTION_CALCULATE = "CALCULATE";
    private JPanel panel;
    
    private JLabel distanceFctLabel;
    private JComboBox distanceFctCombobox;
    private JLabel clusterFctLabel;
    private JComboBox clusterFctCombobox;
    private JLabel thresholdLabel;
    private JCheckBox includeTrivialTinvCheckbox;
    private JCheckBox calculateButton;
    private SpinnerModel model;

    private Project project;

    @Override
    public void run(Project project, ErrorLog log) {
        // Result files.
        File clusterTableFile;
        try {
            clusterTableFile = File.createTempFile("monalisa", ".txt");
        } catch (IOException e) {
            log.log(e.getLocalizedMessage(), ErrorLog.Severity.ERROR);
            e.printStackTrace();
            return;
        }
        clusterTableFile.deleteOnExit();
        
        File clusterTreeFile;
        try {
            clusterTreeFile = File.createTempFile("monalisa", ".xwg");
        } catch (IOException e) {
            log.log(e.getLocalizedMessage(), ErrorLog.Severity.ERROR);
            e.printStackTrace();
            return;
        }
        clusterTreeFile.deleteOnExit();
        
        float percent = ((SpinnerNumberModel) model).getNumber().floatValue() * 100;
        String distanceFunction = getActiveDistanceFunction();
        String clusterAlgorithm = getActiveClusterFunction();
        boolean includeTrivialTInvariants = includeTrivialTinvCheckbox.isSelected();

        if (!computeClusterTree(project, clusterAlgorithm, distanceFunction, clusterTreeFile.getAbsolutePath(), clusterTableFile.getAbsolutePath(), percent, includeTrivialTInvariants)) {
            log.log("#ErrorCompletingClusterToolTreshold", ErrorLog.Severity.ERROR);
            return;
        }
        
        String clusterTableOutput = null;
        try {
            clusterTableOutput = FileUtils.read(clusterTableFile);
        }
        catch (FileNotFoundException e) {
            log.log("#ErrorCompletingClusterTool", ErrorLog.Severity.ERROR);
            e.printStackTrace();
            return;
        }

        String clusterTreeOutput = null;
        try {
            clusterTreeOutput = FileUtils.read(clusterTreeFile);
        }
        catch (FileNotFoundException e) {
            log.log("#ErrorCompletingClusterTool", ErrorLog.Severity.ERROR);
            e.printStackTrace();
            return;
        }

        addResult(new ClusterConfiguration(distanceFunction,
                                          clusterAlgorithm,
                                          percent,
                                          includeTrivialTInvariants, "txt"),
                                          new Clustering(clusterTableOutput, "txt"));

        addResult(new ClusterConfiguration(distanceFunction,
                                          clusterAlgorithm,
                                          percent,
                                          includeTrivialTInvariants, "xwg"),
                                          new Clustering(clusterTreeOutput, "xwg"));

        enableExport(true);
    }
    
    @Override
    public boolean finishedState(Project project) {
        return false;
    }

    @Override
    public JPanel getUI(Project project, StringResources strings) {
        this.project = project;
        if (panel == null) {
            exportButton.setActionCommand(EXPORT);
            exportButton.addActionListener(this);
            enableExport(project.hasResults(this));

            distanceFctLabel = new JLabel(strings.get("DistanceFunction"));
            distanceFctCombobox = new JComboBox();
            
            ComboBoxItem[] distItems = new ComboBoxItem[3];
            int itemIndex = 0;
            for (String df : new String[] {
                    DistanceFunctions.TANIMOTO,
                    DistanceFunctions.SUM_OF_DIFFERENCES,
                    DistanceFunctions.SIMPLEMATCHING })
                distItems[itemIndex++] = new ComboBoxItem(df, DistanceFunctions.getName(df));
            
            distanceFctCombobox.setModel(new DefaultComboBoxModel(distItems));
            
            clusterFctLabel = new JLabel(strings.get("ClusteringFunction"));
            clusterFctCombobox = new JComboBox();
            
            ComboBoxItem[] clusterItems = new ComboBoxItem[5];
            itemIndex = 0;
            for (String cf : new String[] {
                   ClusterFunctions.UPGMA,
                   ClusterFunctions.WPGMA,
                   ClusterFunctions.NeighborJoining,
                   ClusterFunctions.SingleLinkage,
                   ClusterFunctions.CompleteLinkage})
                clusterItems[itemIndex++] = new ComboBoxItem(cf, ClusterFunctions.getName(cf));
            
            clusterFctCombobox.setModel(new DefaultComboBoxModel(clusterItems));
            
            // (similarity of t-invariants grouped in a cluster [%]) 
            thresholdLabel = new JLabel(strings.get("Threshold"));
           
            model = new SpinnerNumberModel(1,     //initial value
                                           0,     //min
                                           1,     //max
                                           0.05); // step
            JSpinner spinner = new JSpinner(model);
            spinner.setEditor(new JSpinner.NumberEditor(spinner, "0.##%"));
          
            includeTrivialTinvCheckbox = new JCheckBox(strings.get("IncludeTrivialTInvariants"));
            includeTrivialTinvCheckbox.setSelected(false);
            
            calculateButton = new JCheckBox(strings.get("Calculate"));
            calculateButton.setActionCommand(ACTION_CALCULATE);
            calculateButton.addActionListener(this);
            
            panel = new JPanel();
            GroupLayout layout = new GroupLayout(panel);
            panel.setLayout(layout);
            
            layout.setAutoCreateGaps(true);
            layout.setAutoCreateContainerGaps(true);
            
            layout.setHorizontalGroup(layout.createParallelGroup()
                .addComponent(distanceFctLabel)
                .addComponent(distanceFctCombobox)
                .addComponent(clusterFctLabel)
                .addComponent(clusterFctCombobox)
                .addGroup(layout.createSequentialGroup()
                    .addComponent(thresholdLabel)
                    .addComponent(spinner))
                .addComponent(includeTrivialTinvCheckbox)
                .addComponent(calculateButton));
//                .addComponent(exportButton));
            
            layout.setVerticalGroup(layout.createSequentialGroup()
                .addComponent(distanceFctLabel)
                .addComponent(distanceFctCombobox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE,
                    GroupLayout.PREFERRED_SIZE)
                .addComponent(clusterFctLabel)
                .addComponent(clusterFctCombobox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE,
                    GroupLayout.PREFERRED_SIZE)
                .addGroup(layout.createParallelGroup(GroupLayout.Alignment.LEADING)
                    .addComponent(thresholdLabel)
                    .addComponent(spinner,  GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE,
                        GroupLayout.PREFERRED_SIZE))
                .addComponent(includeTrivialTinvCheckbox)
                .addComponent(calculateButton));
//                .addComponent(exportButton));
            
            layout.linkSize(SwingConstants.VERTICAL, thresholdLabel, spinner);
        }

        return panel;
    }

    @Override
    public void actionPerformed(ActionEvent e) {
        final String action = e.getActionCommand();

        if (action == ACTION_CALCULATE)
            fireActivityChanged(isActive());
        else if(action == EXPORT) {
            List<Class<? extends Tool>> toolList = new ArrayList<Class<? extends Tool>>();
            toolList.add(ClusterTool.class);
            try {
                this.project.exportResults(toolList);
            } catch (IOException ex) {
                Logger.getLogger(ClusterTool.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
    }

    @Override
    public boolean isActive() {
        return calculateButton.isSelected();
    }

    @Override
    public void saveSettings(Project p) {
        // TODO Auto-generated method stub
    }

    @Override
    public void setActive(boolean active) {
        if (active != isActive()) {
            calculateButton.setSelected(active);
            fireActivityChanged(active);
        }
    }

    @Override
    public void setActive(Configuration... configs) {
        // TODO Auto-generated method stub
        
    }

    @SuppressWarnings("unchecked")
    @Override
    public List<Pair<Class<? extends Tool>, Configuration>> getRequirements() {
        if (isActive())
            return Arrays.asList(
                new Pair<Class<? extends Tool>, Configuration>(
                    TInvariantTool.class, new TInvariantsConfiguration()),
                new Pair<Class<? extends Tool>, Configuration>(
                    DistanceTool.class, new DistanceMatrixConfiguration(
                        getActiveDistanceFunction(),
                        includeTrivialTinvCheckbox.isSelected())));
        else
            return Collections.emptyList();
    }
    
    private String getActiveDistanceFunction() {
        ComboBoxItem selected = (ComboBoxItem) distanceFctCombobox.getSelectedItem();
        return (String) selected.getItem();
    }
    
    private String getActiveClusterFunction() {
        ComboBoxItem selected = (ComboBoxItem) clusterFctCombobox.getSelectedItem();
        return (String) selected.getItem();
    }

    private boolean computeClusterTree(Project project, String clusterFct, String distFct, String target, String output, float threshold, boolean includeTrivialTInv) {
        TInvariants allTinvariants = project.getResult(TInvariantTool.class, new TInvariantsConfiguration());
        List<TInvariant> filteredInvariants = new ArrayList<TInvariant>(includeTrivialTInv ? allTinvariants : allTinvariants.nonTrivialTInvariants());

        DistanceMatrixConfiguration config = new DistanceMatrixConfiguration(distFct, includeTrivialTInv);
        DistanceMatrix matrix = project.getResult(DistanceTool.class, config);
        
        String distanceFunc = DistanceMatrixOutput.getNormalizedString(distFct);

        Cluster c = new Cluster();
        Boolean ret = false;
        try {
            ret = c.generateFile(new ArrayList<TInvariant>(allTinvariants), filteredInvariants, matrix, distanceFunc, clusterFct, target, output, threshold);
            addResult(new ClusterTreeConfiguration(distFct, clusterFct, threshold, includeTrivialTInv, "txt"), new TreeClustering(c.getNodes(), c.getEdges(), c.getGroups()));
        }
        catch(Exception e) {

        }

        return ret;
    }
}
