/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.tools.cluster;

import java.io.*;
import java.util.*;

import monalisa.data.pn.PetriNet;
import monalisa.data.pn.TInvariant;
import monalisa.data.pn.Transition;

/**
 * Die Klasse Group.java implementiert die Generierung der output.txt-Datei
 * inklusive der Berechnung der MCT-sets.
 * 
 * @author belau
 */
public class Group {

    /** Creates a new instance of Group */
    public Group() {

    }

    /** Methode zur Erzeugung der output.txt-Datei */
    public void generateGroupFile(
            float matrix[][], 
            String[] label,
            List<TInvariant> allTinvariants, 
            List<TInvariant> filteredTinvariants,
            List<Leaves> groups, 
            String outputfile) {

        /**  Die  output.txt-Datei enthaelt u.a. Informationen darueber, welche Transitionen/MCT-sets 
         *   von den T-Invarianten eines Clusters ueberdeckt werden. Dies wird in Form einer Tabelle
         *   dargestellt:
         *                  T1  T2  MCTS1  T3 ...          (title-Vector)
         *   Cluster1 (3):  1   2   *                   (l.resultVector: 1,2 *, , ...)
         *   Cluster2 (3):  *   1       *               (l.resultVector)
         *   Cluster3 (2):      *   1                   (l.resultVector)
         *   
         *   Die Zahlen geben an, wieviele T-Inv. eines Clusters eine Transition/MCT-set ueberdecken, * gibt an,
         *   dass alle T-Inv. eines Clusters die  Transition/MCT-set ueberdecken.
         *   
         *   Jedes Blaettercluster ist durch ein Objekt l der Klasse Leaves wiedergegeben. Die in dem Vector groups
         *   abgespeicherten Objekte enthalten als Objektattribute:
         *   - name: Blaettercluster-Label (String)
         *   - member: T-Inv.-Label (bsp: Inv_1) der in einem Blaettercluster enthaltenen Invarianten (String)
         *   - label: "Cluster"+Blaettercluster-Label (String) 
         *  
         *   - lmember:  Ids der T-Inv. eines Blaettercluster (Vector) -> wird in dieser Methode "berechnet"
         *   - memberNo: Anzahl der T-Inv. in einem Blaettercluster (int) -> wird in dieser Methode berechnet
         *   - result: resultVector (siehe oben) -> wird in dieser Methode berechnet
         *   
         */  

        // Berechnung der Objektattribute lmember, memberNo und result fuer jedes Blaettercluster l aus groups

        for (int i = 0; i < groups.size(); i++) {
            ArrayList<String> tinvForOneCluster = new ArrayList<String>();
            Leaves l = groups.get(i);
            String tinvs = l.member;
            // split comma-separated Inv-String
            StringTokenizer s = new StringTokenizer(tinvs);
            while (s.hasMoreTokens()) {
                String st = s.nextToken();
                st = st.substring(0, st.length() - 1);
                tinvForOneCluster.add(st);
            }
            // Initialisierung memberNo
            l.memberNo = tinvForOneCluster.size();
            // Initialisierung lmember
            for (int j = 0; j < tinvForOneCluster.size(); j++) {
                String nameInv = tinvForOneCluster.get(j);
                int idInv = Integer.parseInt(nameInv.replace("Inv", ""));
                l.lmember.add(idInv);
            }
            // Berechnung des resultVectors
            l.result = generateResultVector(allTinvariants, tinvForOneCluster);
        }

        /** interne Clustervalidierung */
        Validation v = new Validation();
        v.validate(matrix, label, groups);

        /** Berechnung der MCT-sets. */
        Vector<String> mctsVec = generateMctSetResultVector(filteredTinvariants);

        /**
         * Generierung der Tabelle (Beispiel: siehe oben). Dazu ist es u.a.
         * notwendig, die in einem MCT-set enthaltenen Transitionen
         * zusammenzufassen: -> Abgleich des title-vectors mit MCT-sets
         * -> Abgleich der result-Vectoren mit den MCT-sets Generierung von 2
         * ArrayLists: 1. delete: enthaelt die zu loeschende Transitionen, 2.
         * substitute: enhaelt die zu ersetzende Transitionen durch MCT-sets
         */

        ArrayList<String> substitute = new ArrayList<String>();
        ArrayList<String> delete = new ArrayList<String>();
        
        for (int j = 0; j < mctsVec.size(); j++) {
            ArrayList<String> line = new ArrayList<String>();
            String mcts = mctsVec.get(j);
            StringTokenizer s = new StringTokenizer(mcts);
            while (s.hasMoreTokens()) {
                line.add(s.nextToken());
            }
            // Example for set: "MCTS 1 (MCT1): T0, T1, T6, T7, T8, T10, T30;"
            for (int i = 1; i < line.size(); i++) {
                if (i == 1) {
                    String MCT = line.get(i);
//                    MCT = MCT.substring(5, MCT.length() - 2);
                    substitute.add(MCT);
                }
                if (i > 2) {
                    String T = line.get(i);
                    T = T.substring(1, T.length() - 1);
                    delete.add(T);
                }
            }
        }

        /* Spaltenbeschriftung der Tabelle (title-vector) */
        TInvariant t1 = filteredTinvariants.get(0);

        // Get the sorted transitions.
        Iterator<Transition> it = t1.iterator();
        PetriNet petriNet = it.next().container();
        List<Transition> sortedTransitions = new ArrayList<Transition>(petriNet.transitions());
        Collections.sort(sortedTransitions);
        // END Get the sorted transitions.

        int firstId = sortedTransitions.get(0).id();
        int id = firstId;
        Vector<String> title = new Vector<String>();
        title.add("Transition/MCTS: ");
        for (int i = 0; i < sortedTransitions.size(); i++) {
            title.add("T" + (firstId++) + ",");
        }

        /* Abgleich des title-vectors mit MCT-Sets */
        for (int i = 0; i < delete.size(); i++) {
            int b = Integer.parseInt(delete.get(i));
            if (id == 0) {
                title.setElementAt("", b + 1);
            } else {
                title.setElementAt("", b);
            }
        }

        int no = 1;
        for (int k = 0; k < substitute.size(); k++) {
            String name = "MCTS" + no + ",";
            int d = Integer.parseInt(substitute.get(k));
            if (id == 0) {
                title.setElementAt(name, d + 1);
            } else {
                title.setElementAt(name, d);
            }
            no++;
        }

        /* Abgleich der result-Vectoren mit den MCT-Sets */
        for (int i = 0; i < groups.size(); i++) {
            Leaves l = groups.get(i);
            for (int k = 0; k < delete.size(); k++) {
                int b = Integer.parseInt(delete.get(k));
                if (id == 0) {
                    l.result.setElementAt("", b);
                } else {
                    l.result.setElementAt("", b - 1);
                }
            }
        }

        /** Generierung der output.txt-Datei */
        try {
            PrintWriter writer = new PrintWriter(new FileWriter(outputfile));
            
            writer.println("Elementary modes grouped in a cluster");
            writer.println("The number of elementary modes in a cluster is put in brackets.");
            
            for (int i = 0; i < groups.size(); i++) {
                Leaves l = groups.get(i);
                writer.print(l.label + "(" + l.memberNo + "): ");
                writer.println(l.member);
            }

            
//            writer.println("TRANSITION(T)/MCTSET(MCTS) COVERED BY INVARIANTS GROUPED IN A CLUSTER");
//
//            // Beschriftung der 1. Zeile der Tabelle
//            for (int j = 0; j < title.size(); j++) {
//                writer.print(title.get(j));
//            }
//
//            writer.println();
//
//            for (int i = 0; i < groups.size(); i++) {
//                Leaves l = groups.get(i);
//                writer.print(l.label + " (" + l.memberNo + "):\t");
//                for (int j = 0; j < l.result.size(); j++) {
//                    writer.print(l.result.get(j));
//                }
//                writer.println();
//            }
//            
//            writer.println();
//            writer.println(" *:    the Transition/MCTS is covered by all invariants of a cluster");
//            writer.println(" .:    the Transition/MCTS is covered by none of the invariants of a cluster");
//            writer.println(" no.:  the Transition/MCTS is covered by this number of invariants of a cluster");
//            writer.println();
//            writer.println();
//            writer.println("MCTS (maximal common transitions set):");
//
//            for (int j = 0; j < mctsVec.size(); j++) {
//                writer.println(mctsVec.get(j) + " ");
//            }
//
//            writer.println();
            
            writer.close();
            
        } catch (IOException ex) {
            ex.printStackTrace();
            System.out.println("Fehler bei Erzeugung der output.txt-Datei");
        }
    }

    /**
     * Methode zur Berechnung des resultVectors eines Blaetterclusters (siehe oben).
     */
    public static Vector<String> generateResultVector(
            List<TInvariant> allTinvariants, 
            ArrayList<String> list) {
        // list = alle invarianten eines clusters?
        /**
         * Es wird der tInvariant-Vector inklusive triviale T-Inv. benoetigt, da
         * in diesem der Index der Invarianten-Id-1 entspricht: Auf die
         * entsprechende Inv kann somit ohne Suchfunktion direkt zugegriffen
         * werden
         */
        
        Vector<String> vectorRes = new Vector<String>();
        /* Deklaration/Initialisierung von vectorResult (enthaelt Haufigkeiten) */
        TInvariant t = allTinvariants.get(0);
        PetriNet petriNet = t.iterator().next().container();
        int size = petriNet.transitions().size(); // Groeße support-Vector

        int[] vectorResult = new int[size];
        for (int k = 0; k < vectorResult.length; k++) {
            vectorResult[k] = 0;
        }

        for (int i = 0; i < list.size(); i++) {
            String nameInv = list.get(i);
            int idInv = Integer.parseInt(nameInv.replace("Inv", ""));
            TInvariant t1 = allTinvariants.get(idInv);
            List<Integer> vectorInv1 = t1.asVector();
            for (int k = 0; k < vectorInv1.size(); k++) {
                int a = vectorResult[k];
                if (vectorInv1.get(k) > 0) {
                    a += 1;
                }
                vectorResult[k] = a;
            }
        }

        /**
         * Initialisierung von vectorRes (enthaelt Haufigkeiten codiert als *; .
         * ; Zahhlen )
         */
        int s = list.size();
        for (int k = 0; k < size; k++) {
            int a = vectorResult[k];
            if (a == s) { // alle T-Inv. eines Clusters ueberdecken die
                // Transition/MCT-set
                vectorRes.add("*,");
            } else if (a == 0) { // keine der T-Inv. eines Clusters ueberdeckt
                // die Transition/MCT-set
                vectorRes.add(".,");
            } else { // a T-Inv. eines Clusters ueberdecken die
                // Transition/MCT-set
                vectorRes.add(a + ",");
            }
        }
        return vectorRes;
    }

    /**
     * Methode zur Berechnung der MCT-sets. Die MCT-sets werden auf der
     * Grundlage eines String-Vergleichs berechnet, wobei fuer jede Transition
     * ein String deklariert wird, der die Ids derjenigen T-inv. auflistet, in
     * der diese Transition enthalten ist: t1: 157 (157: Id-String der
     * Transition 1) t2: 23 t3: 157 .... -> MCT-set: t1, t3 Durch den
     * String-Vergleich kann ermittelt werden, welche Transitionen in den
     * gleichen T-Inv. vorkommen bzw. welche Transitionen ein MCT-set bilden
     */

    public static Vector<String> generateMctSetResultVector(
            List<TInvariant> tInvariants) {

        /**
         * Deklaration/Initialisierung des String-arrays tSet. Dieser enthaelt
         * die "Id-Strings" der einzelnen Transitionen (siehe oben)
         */

        TInvariant t = tInvariants.get(0);
        PetriNet petriNet = t.iterator().next().container();
        int size = petriNet.transitions().size();
        String[] tSet = new String[size];

        for (int k = 0; k < size; k++) {
            tSet[k] = "";
        }

        for (int i = 0; i < tInvariants.size(); i++) {
            TInvariant t1 = tInvariants.get(i);
            Integer id = t1.id();
            String name = id.toString();
            List<Integer> vectorInv1 = t1.asVector();
            for (int k = 0; k < vectorInv1.size(); k++) {
                String a = tSet[k];

                if (vectorInv1.get(k) > 0) {
                    a = a + name;
                }
                tSet[k] = a;
            }
        }

        /**
         * Deklaration/Initialisierung des Vectors tSET. Dieser enthaelt die
         * berechneten MCT-sets als Strings (bsp: MCT-set 1 (T1): t1, t3, t4...)
         */

        Vector<String> mcts = new Vector<String>();

        // Get the sorted transitions.
        List<Transition> sortedTransitions = new ArrayList<Transition>(petriNet.transitions());
        Collections.sort(sortedTransitions);
        // END Get the sorted transitions.

        int firstId = sortedTransitions.get(0).id();
        int counter = 1;
        int counter1 = 0;
        for (int i = 0; i < tSet.length; i++) {

            if (tSet[i] == "n") { // alle Transitionen, die bereits in einem
                // MCT-set enthalten sind werden bei dem
                // Vergleich nicht mehr berueckstichtigt
                // Empty!!
            } else {
                String set = "MCTS ";

                // Unterscheidung, ob triviale bzw. nicht-triviale T-Inv. zur
                // MCT-set-Berechnung verwendet werden
                if (firstId == 0) {
                    set = set + counter + " (MCTS" + counter + "): " + "T" + (i);
                } else {
                    set = set + counter + " (MCTS" + counter + "): " + "T"
                            + (i + 1);
                }

                for (int j = i + 1; j < tSet.length; j++) {
                    if (tSet[j] == "n") {
                        // Empty!!
                    } else {
                        String a = tSet[i];
                        String b = tSet[j];

                        if (a.compareTo(b) == 0) { // String-Vergleich
                            ++counter1;
                            if (firstId == 0) {
                                set = set + ", T" + (j);
                            } else {
                                set = set + ", T" + (j + 1);
                            }
                            tSet[j] = "n"; // Markierung der Transitionen, die
                            // bereits verglichen wurden
                        }
                    }
                }
                set = set + ";";
                if (counter1 > 0) { // nur MCT-sets mit mehr als einer
                    // Transition werden beruecksichtigt
                    ++counter;
                    mcts.add(set);
                }
                counter1 = 0;
            }
        }
        return mcts;
    }
}